import os
from math import floor
from pathlib import Path
from textwrap import dedent
from typing import List, Optional

import typer
from pynmrstar import Entry
from tabulate import tabulate

from nef_pipelines.lib.nef_lib import (
    SelectionType,
    read_or_create_entry_exit_error_on_bad_file,
    select_frames,
)
from nef_pipelines.lib.sequence_lib import count_residues, frame_to_chains
from nef_pipelines.lib.typer_utils import get_args
from nef_pipelines.lib.util import chunks, exit_error
from nef_pipelines.tools.frames import frames_app

UNDERSCORE = "_"

parser = None


def _if_is_nef_file_load_as_entry(file_path):
    entry = None
    try:
        entry = Entry.from_file(file_path)
    except Exception:
        pass

    return entry


# noinspection PyUnusedLocal
@frames_app.command()
def list(
    pipe: Path = typer.Option(
        None,
        "-i",
        "--in",
        metavar="NEF-FILE",
        help="read NEF data from a file instead of stdin",
    ),
    selector_type: SelectionType = typer.Option(
        SelectionType.ANY,
        "-t",
        "--selector-type",
        help=f"force how to select frames can be one of {', '.join(SelectionType.__members__)}",
    ),
    number: bool = typer.Option(False, "-n", "--number", help="number entries"),
    verbose: int = typer.Option(
        False,
        "-v",
        "--verbose",
        count=True,
        help="print verbose information more verbose options give more information",
    ),
    filters: Optional[List[str]] = typer.Argument(
        None, help="filters string for entry names and categories to list"
    ),
):
    """- list the frames in the current input"""

    entry = None
    if len(filters) > 0:
        entry = _if_is_nef_file_load_as_entry(filters[0])
        if entry is not None:
            if pipe:
                msg = f"""\
                   two nef file paths supplied...
                       path 1: {pipe} [from --in]
                       path 2: {filters[0]} [from args]"""
                msg = dedent(msg)
                exit_error(msg)
            else:
                pipe = filters[0]
                filters = filters[1:]
    if not filters:
        filters = [
            "*",
        ]

    args = get_args()

    if entry is None:

        try:
            entry = read_or_create_entry_exit_error_on_bad_file(pipe)
        except Exception as e:
            exit_error(f"failed to read nef file {args.pipe} because", e)
        lines = str(entry)

    if entry is None:
        if args.pipe is not None:
            exit_error(f"couldn't read a nef stream from the file: {args.pipe}")
        elif len(filters) > 0:
            exit_error(f"couldn't read a nef stream from either {filters[0]} or stdin")
        else:
            exit_error("couldn't read a nef stream from stdin")

    print(f"entry {entry.entry_id}")
    if verbose:

        import hashlib

        md5 = hashlib.md5(lines.encode("ascii")).hexdigest()
        num_lines = len(lines.split("\n"))
        print(f"    lines: {num_lines} frames: {len(entry)} checksum: {md5} [md5]")
    print()

    frames = select_frames(entry, selector_type, filters)

    if verbose == 0:
        frame_names = [frame.name for frame in frames]
        if number:
            frame_names = [
                f"{i}. {frame_name}"
                for i, frame_name in enumerate(frame_names, start=1)
            ]

        frame_list = _string_list_to_tabulation(frame_names)
        print(tabulate(frame_list, tablefmt="plain"))

    else:

        for frame in enumerate(frames, start=1):

            filters = [f"*{filter}*" for filter in filters]

            if verbose == 0:
                frame_names = [
                    f"{i}. {frame.name}" for i, frame_name in enumerate(frames, start=1)
                ]

                frame_list = _string_list_to_tabulation(frame_names)
                print(tabulate(frame_list, tablefmt="plain"))

                print(f"    category: {frame.category}")
                if len(frame.name) != len(frame.category):
                    print(
                        f"    name: {frame.name[len(frame.category):].lstrip(UNDERSCORE)}"
                    )

                loop_lengths = []
                for loop in frame.loop_dict:
                    loop_lengths.append(str(len(loop)))

                loops = ""
                if len(loops) == 1:
                    loops = " [length: 1]"
                else:
                    loops = f' [lengths: {", ".join(loop_lengths)}]'

                print(f"    loops: {len(frame.loop_dict)}{loops}")

                frame_standard = frame.name[: len("nef")]
                is_standard_frame = frame_standard == "nef"
                print(f"    is nef frame: {is_standard_frame}")

                # ccpn_compound_name
                if verbose == 2 and frame.category == "nef_molecular_system":
                    chains = frame_to_chains(frame)

                    print(f'    chains: {len(chains)} [{", ".join(chains)}]')

                    residue_counts = {}
                    for chain in chains:
                        residue_counts[chain] = count_residues(frame, chain)

                    residue_count_per_chain = {}
                    for chain in chains:
                        residue_count_per_chain[chain] = sum(
                            residue_counts[chain].values()
                        )

                    output = [
                        f"{chain} {num_residues}"
                        for chain, num_residues in residue_count_per_chain.items()
                    ]
                    print(f'    residues: {", ".join(output)}')

                    for chain in chains:
                        counts_and_percentages = []

                        for residue, count in residue_counts[chain].items():
                            percentage = (
                                f"{count/ residue_count_per_chain[chain]*100:5.2f}"
                            )
                            counts_and_percentages.append(
                                f"{residue}: {count} [{percentage}%]"
                            )

                        pre_string = f"              {chain}. "
                        pre_string_width = len(pre_string)

                        tabulation = _string_list_to_tabulation(
                            counts_and_percentages, pre_string_width
                        )
                        table = tabulate(tabulation, tablefmt="plain")

                        print(_indent_with_prestring(table, pre_string))

    print()


def _indent_with_prestring(text_block, pre_string):
    raw_result = []
    empty_prestring = " " * len(pre_string)
    for i, string in enumerate(text_block.split("\n")):
        if i == 0:
            raw_result.append(f"{pre_string}{string}")
        else:
            raw_result.append(f"{empty_prestring}{string}")

    return "\n".join(raw_result)


def _string_list_to_tabulation(frame_names, used_columns=0):
    try:
        width, _ = os.get_terminal_size()
    except Exception:
        width = 100

    width -= used_columns

    # apply a sensible minimum width
    if width < 20:
        width = 20

    if len(frame_names) > 0:
        frame_name_widths = [len(frame_name) for frame_name in frame_names]
        max_frame_name_width = max(frame_name_widths)

        columns = int(floor(width / (max_frame_name_width + 1)))
        column_width = int(floor(width / columns))

        columns = 1 if columns == 0 else columns

        frame_names = [frame_name.rjust(column_width) for frame_name in frame_names]
        frame_name_list = chunks(frame_names, columns)
    else:
        frame_name_list = [[]]

    return frame_name_list



from nef_pipelines.lib.test_lib import select_matching_tests

TEST_NAMES = [
    "tests/nmrpipe/test_gdb.py::test_bad_data_format",
    "tests/nmrpipe/test_gdb.py::test_bad_format",
    "tests/nmrview/test_tcl.py::test_basic_word",
    "tests/nmrview/test_tcl.py::test_quoted_word",
]


def test_test_match_specific():

    EXPECTED_NAMES = [
        "tests/nmrpipe/test_gdb.py::test_bad_data_format",
        "tests/nmrview/test_tcl.py::test_basic_word",
    ]

    result = select_matching_tests(TEST_NAMES, EXPECTED_NAMES)

    assert result == EXPECTED_NAMES


def test_test_match_stared_path():

    EXPECTED_NAMES = [
        "tests/nmrview/test_tcl.py::test_basic_word",
    ]

    args = ["*nmrview*::test_basic_word"]

    result = select_matching_tests(TEST_NAMES, args)

    assert result == EXPECTED_NAMES


def test_test_match_stared_name():
    args = ["tests/nmrview/test_tcl.py::*"]

    EXPECTED_NAMES = [
        "tests/nmrview/test_tcl.py::test_basic_word",
        "tests/nmrview/test_tcl.py::test_quoted_word",
    ]

    result = select_matching_tests(TEST_NAMES, args)

    assert result == EXPECTED_NAMES


def test_just_name():
    args = ["test_basic_word"]

    EXPECTED_NAMES = [
        "tests/nmrview/test_tcl.py::test_basic_word",
    ]

    result = select_matching_tests(TEST_NAMES, args)

    assert result == EXPECTED_NAMES


def test_star():
    args = ["*"]

    result = select_matching_tests(TEST_NAMES, args)

    assert result == TEST_NAMES


def test_test_match_truncated_path():

    EXPECTED_NAMES = [
        "tests/nmrpipe/test_gdb.py::test_bad_data_format",
        "tests/nmrpipe/test_gdb.py::test_bad_format",
    ]

    args = ["nmrpipe/test_gdb.py::*"]
    result = select_matching_tests(TEST_NAMES, args)

    assert result == EXPECTED_NAMES


def test_no_path():
    args = ["::test_bad_format"]

    EXPECTED_NAMES = [
        "tests/nmrpipe/test_gdb.py::test_bad_format",
    ]

    result = select_matching_tests(TEST_NAMES, args)

    assert result == EXPECTED_NAMES


def test_no_name():
    args = ["tests/nmrpipe/test_gdb.py::"]

    EXPECTED_NAMES = [
        "tests/nmrpipe/test_gdb.py::test_bad_data_format",
        "tests/nmrpipe/test_gdb.py::test_bad_format",
    ]

    result = select_matching_tests(TEST_NAMES, args)

    assert result == EXPECTED_NAMES


def test_bare_separator():

    args = ["::"]

    result = select_matching_tests(TEST_NAMES, args)

    assert result == TEST_NAMES


def test_bare_filename():

    args = ["test_gdb.py::"]

    EXPECTED_NAMES = [
        "tests/nmrpipe/test_gdb.py::test_bad_data_format",
        "tests/nmrpipe/test_gdb.py::test_bad_format",
    ]

    result = select_matching_tests(TEST_NAMES, args)

    assert result == EXPECTED_NAMES



# TODO add support for DATA FIRST_RESID
# noinspection PyUnusedLocal
from argparse import Namespace
from pathlib import Path
from typing import Iterable, List

import typer

from nef_pipelines.lib.sequence_lib import (
    chain_code_iter,
    offset_chain_residues,
    sequence_3let_to_sequence_residues,
    sequence_to_nef_frame,
)
from nef_pipelines.lib.structures import SequenceResidue
from nef_pipelines.lib.typer_utils import get_args
from nef_pipelines.lib.util import (
    cached_file_stream,
    exit_error,
    process_stream_and_add_frames,
)
from nef_pipelines.transcoders.nmrpipe import import_app
from nef_pipelines.transcoders.nmrpipe.nmrpipe_lib import (
    gdb_to_3let_sequence,
    read_db_file_records,
)

app = typer.Typer()


# noinspection PyUnusedLocal
@import_app.command()
def sequence(
    chain_codes: List[str] = typer.Option(
        None,
        "--chains",
        help="chain codes as a list of names separated by dots",
        metavar="<CHAIN-CODES>",
    ),
    no_chain_start: bool = typer.Option(
        False,
        "--no-chain-start/",
        help="don't include a start of chain link type for the first residue",
    ),
    no_chain_end: bool = typer.Option(
        False,
        "--no-chain-end/",
        help="don't include an end of chain link type for the last residue",
    ),
    entry_name: str = typer.Option("nmrpipe", help="a name for the entry"),
    pipe: Path = typer.Option(
        None,
        metavar="|PIPE|",
        help="pipe to read NEF data from, for testing [overrides stdin !use stdin instead!]",
    ),
    start: int = typer.Option(
        1, metavar="offset", help="residue number to start the sequence at"
    ),
    file_names: List[Path] = typer.Argument(
        ..., help="input files of type nmrview.seq", metavar="<SEQ-FILE>"
    ),
):
    """convert nmrview sequence file <nmrview>.seq files to NEF"""
    try:
        args = get_args()

        process_sequence(args)
    except Exception as e:
        exit_error(f"reading sequence failed because {e}", e)


def process_sequence(args: Namespace):
    nmrpipe_frames = []

    chain_codes = args.chain_codes
    if not chain_codes:
        chain_codes = ["A"]
    for file_name, chain_code in zip(args.file_names, chain_code_iter(chain_codes)):
        # cached_file_stream
        with cached_file_stream(file_name) as lines:

            nmrpipe_sequence = read_sequence(
                lines,
                chain_code=chain_code,
                sequence_file_name=file_name,
                start=args.start,
            )

            frame = sequence_to_nef_frame(nmrpipe_sequence, args.entry_name)

            nmrpipe_frames.append(frame)

    entry = process_stream_and_add_frames(nmrpipe_frames, args)

    print(entry)


def read_sequence(
    sequence_lines: Iterable[str],
    chain_code: str = "A",
    sequence_file_name: str = "unknown",
    start: int = 1,
) -> List[SequenceResidue]:

    gdb_file = read_db_file_records(sequence_lines, sequence_file_name)

    sequence_3let = gdb_to_3let_sequence(gdb_file)

    sequence_residues = sequence_3let_to_sequence_residues(
        sequence_3let, chain_code=chain_code
    )

    sequence_residues = offset_chain_residues(
        sequence_residues, {chain_code: start - 1}
    )

    return sequence_residues


if __name__ == "__main__":

    typer.run(sequence)



import fnmatch
import sys
from collections import Counter
from pathlib import Path
from typing import Dict, List

import typer
from pynmrstar import Saveframe
from tabulate import tabulate as tabulate_formatter
from tabulate import tabulate_formats

from nef_pipelines.lib.nef_lib import (
    UNUSED,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.typer_utils import get_args
from nef_pipelines.lib.util import exit_error, read_integer_or_exit
from nef_pipelines.tools.frames import frames_app

ALL_LOOPS = "ALL"

UNDERSCORE = "_"

parser = None

EXIT_ERROR = 1

FORMAT_HELP = f'format for the table, [possible formats are: {", ".join(tabulate_formats)}: those provided by tabulate]'
FRAMES_HELP = (
    "selectors for loops to tabulate. note: wild cards are allowed and specific loops can be chosen"
    " (see above)"
)

ABBREVIATED_HEADINGS = {
    "index": "ind",
    "chain_code": "chain",
    "sequence_code": "seq",
    "residue_name": "resn",
    "linking": "link",
    "cis_peptide": "cis",
    "ccpn_compound_name": "ccpn_name",
    "ccpn_chain_role": "ccpn_role",
    "ccpn_comment": "comment",
    "ccpn_dataset_serial": "data_serial",
    "program_name": "name",
    "program_version": "version",
    "script_name": "script",
    "peak_id": "id",
    "volume": "vol",
    "volume_uncertainty": "vol-err",
    "height_uncertainty": "height-err",
    "position": "pos",
    "position_uncertainty": "pos-err",
    "ccpn_figure_of_merit": "merit",
    "ccpn_annotation": "ann",
    "ccpn_peak_list_serial": "ccpn-serial",
}


# noinspection PyUnusedLocal
@frames_app.command(no_args_is_help=True)
def tabulate(
    pipe: Path = typer.Option(
        None,
        metavar="|PIPE|",
        help="pipe to read NEF data from, for testing [overrides stdin !use stdin instead!]",
    ),
    format: str = typer.Option(
        "plain", "-f", "--format", help=FORMAT_HELP, metavar="format"
    ),
    verbose: bool = typer.Option(
        False, "-v", "--verbose", help="include frame names etc in output"
    ),
    full: bool = typer.Option(
        False, "--full", help="don't suppress empty columns [default: False]"
    ),
    no_abbreviations: bool = typer.Option(
        False, "--no-abbreviations", help="dont' abbreviate column headings"
    ),
    frame_selectors: List[str] = typer.Argument(
        None, help=FRAMES_HELP, metavar="<loop-selector>..."
    ),
):
    """- tabulate loops from frames in a NEF file.     notes: 1. using the name of a frame will tabulate all loops,
    using frame.loop_index [e.g. moleculecular_system.1] will tabulate a specific loop. 2. wild cards can be use for
    frame names e.g. mol would select molecular_system" and anything other frame whose name contains mol 3. by default
    empty columns are ignored"""

    args = get_args()

    entry = read_or_create_entry_exit_error_on_bad_file(pipe)

    tabulate_frames(entry, args)


def tabulate_frames(entry, args):
    frames_to_tabulate = _get_frames_to_tabulate(entry, args.frame_selectors)
    frame_indices = _select_loop_indices(args.frame_selectors)

    for frame_name, frame_data in frames_to_tabulate.items():
        category = frame_data.category

        category_length = len(category)

        frame_id = frame_data.name[category_length:].lstrip("_")
        frame_id = frame_id.strip()
        frame_id = frame_id if len(frame_id) > 0 else ""

        for i, loop in enumerate(frame_data.loops, start=1):
            if _should_output_loop(i, frame_name, frame_indices):
                _output_loop(loop, frame_id, category, args)


def _select_loop_indices(frames):
    frame_indices = {}
    for frame_selector in frames:
        frame_selector_index = frame_selector.split(".")
        if len(frame_selector_index) == 1:
            frame_name_selector = frame_selector_index[0]
            frame_indices.setdefault(frame_name_selector, set()).add(ALL_LOOPS)
        elif len(frame_selector_index) == 2:
            frame_name_selector = frame_selector_index[0]
            index_string = frame_selector_index[1]
            message = f"[frames tabulate] expected an index got '{index_string}' from {'.'.join(frame_selector_index)}"
            index = read_integer_or_exit(index_string, message=message)
            frame_indices.setdefault(frame_name_selector, set()).add(index)
        else:
            exit_error(
                f'[frames tabulate] too many fields int the frame selector {".".join(frame_selector)}'
            )
    for frame_selector, indices in frame_indices.items():
        if ALL_LOOPS in indices and len(indices) > 1:
            indices.remove(ALL_LOOPS)
            index_strings = ", ".join([str(index) for index in indices])
            message = (
                "[frames tabulate] incompatible selections, you selected all loops and a specific index for "
                + +f"frame: {frame_selector} indices were: {index_strings}"
            )
            exit_error(message)

    return frame_indices


def _remove_empty_columns(tabulation, headers):

    counter = Counter()
    for row in tabulation:
        for i, header in enumerate(headers):
            if row[i] != UNUSED:
                counter[i] += 1

    empty_columns = []
    for column_index, heading in enumerate(headers):
        if not counter[column_index]:
            empty_columns.append(column_index)

    empty_columns = list(reversed(empty_columns))

    for row in tabulation:
        for column_index in empty_columns:
            del row[column_index]

    for column_index in empty_columns:
        del headers[column_index]

    return tabulation, headers


def _output_loop(loop, frame_id, category, args):

    if args.verbose:
        print()
        if frame_id:
            print(f"{frame_id}: {category}/{loop.category[1:]}")
        else:
            print(f"{category}/{loop.category[1:]}")
        print()
    table = []
    headers = loop.tags
    for line in loop.data:
        row = list(line)
        table.append(row)

    if not args.full:
        table, headers = _remove_empty_columns(table, headers)

    if not args.no_abbreviations:
        headers = _abbreviate_headers(headers, ABBREVIATED_HEADINGS)

    print(tabulate_formatter(table, headers=headers, tablefmt=args.format))


def _abbreviate_headers(headers: List[str], abbreviations: Dict[str, str]) -> List[str]:

    result = []

    for header in headers:
        for stem in abbreviations:

            if header.startswith(stem):

                header = f"{abbreviations[stem]}{header[len(stem):]} "

        result.append(header.replace("_", "-"))

    return result


def _should_output_loop(index, frame_name, frame_indices):
    do_output = True
    for frame_selector in frame_indices:
        if fnmatch.fnmatch(frame_name, f"*{frame_selector}*"):
            indices = frame_indices[frame_selector]
            if index not in indices and ALL_LOOPS not in indices:
                do_output = False
    return do_output


def _count_loop_rows(save_frame: Saveframe) -> int:
    count = 0
    for loop in save_frame.loops:
        count += len(loop)

    return count


def _get_frames_to_tabulate(entry, frame_selectors):

    frames_with_empty_loops = _get_loops_with_empty_frames_and_errors(entry)

    frames_to_tabulate = _select_chosen_frames(entry, frame_selectors)

    _remove_empty_frames_and_warn(frames_to_tabulate, frames_with_empty_loops)

    return frames_to_tabulate


def _remove_empty_frames_and_warn(frames_to_tabulate, frames_with_empty_loops):
    for frame_name in frames_to_tabulate:
        if frame_name in frames_with_empty_loops:
            print(frames_with_empty_loops[frame_name], file=sys.stderr)
    for frame_name in frames_with_empty_loops:
        if frame_name in frames_to_tabulate:
            del frames_to_tabulate[frame_name]


def _select_chosen_frames(entry, frame_selectors):
    frames_to_tabulate = {}
    for frame_name, data in entry.frame_dict.items():
        if len(frame_selectors) > 0:
            for frame_selector in frame_selectors:
                frame_selector = frame_selector.split(".")[0]

                if fnmatch.fnmatch(frame_name, f"*{frame_selector}*"):
                    frames_to_tabulate[frame_name] = data
        else:

            frames_to_tabulate[frame_name] = data
    return frames_to_tabulate


def _get_loops_with_empty_frames_and_errors(entry):
    frames_with_empty_loops = {}
    for frame_name, data in entry.frame_dict.items():
        if len(data.loops) == 0:
            msg = f"WARNING: frame {frame_name} has no loops and the frame will be ignored"
            frames_with_empty_loops[frame_name] = msg

        if frame_name not in frames_with_empty_loops and _count_loop_rows(data) == 0:
            msg = f"WARNING: frame {frame_name} has loops but they contain no data and the frame will be ignored"
            frames_with_empty_loops[frame_name] = msg
    return frames_with_empty_loops



import io
from textwrap import dedent

import pytest

from nef_pipelines.lib.structures import LineInfo
from nef_pipelines.tests.test_sequence_lib import ABC_SEQUENCE_1LET, ABC_SEQUENCE_3LET
from nef_pipelines.transcoders.nmrpipe.nmrpipe_lib import (
    VALUES,
    BadFieldFormat,
    DataBeforeFormat,
    DbFile,
    DbRecord,
    MultipleFormat,
    MultipleVars,
    NoVarsLine,
    WrongColumnCount,
    gdb_to_3let_sequence,
    get_column_indices,
    get_gdb_columns,
    read_db_file_records,
    select_records,
)


def _lines_to_line_info(lines):

    result = []

    lines = [line for line in lines.split("\n")]
    for line_no, line in enumerate(lines, start=1):
        if len(line.strip()) != 0:
            result.append(
                LineInfo(file_name="unknown", line_no=line_no, line=f"{line}\n")
            )
    return result


def test_read_gdb_file():
    PALES_DATA = """\
    DATA SEQUENCE MQIFVKTLTG KTITLEVEPS DTIENVKAKI QDKEGIPPDQ QRLIFAGKQL
    DATA SEQUENCE EDGRTLSDYN IQKESTLHLV LRLRGG

    VARS   RESID_I RESNAME_I ATOMNAME_I RESID_J RESNAME_J ATOMNAME_J D      DD    W
    FORMAT %5d     %6s       %6s        %5d     %6s       %6s    %9.3f   %9.3f %.2f

        2    GLN      N      2    GLN     HN     -15.524     1.000 1.00
        3    ILE      N      3    ILE     HN      10.521     1.000 1.00
        4    PHE      N      4    PHE     HN       9.648     1.000 1.00
        5    VAL      N      5    VAL     HN       6.082     1.000 1.00

        1    MET      C      2    GLN     HN       3.993     0.333 3.00
        2    GLN      C      3    ILE     HN      -5.646     0.333 3.00
        3    ILE      C      4    PHE     HN       1.041     0.333 3.00
        4    PHE      C      5    VAL     HN       0.835     0.333 3.00
    """

    pales_stream = io.StringIO(PALES_DATA)

    result = read_db_file_records(pales_stream)

    line_info_items = _lines_to_line_info(PALES_DATA)

    base_records = [
        (
            1,
            "DATA",
            [
                "SEQUENCE",
                "MQIFVKTLTG",
                "KTITLEVEPS",
                "DTIENVKAKI",
                "QDKEGIPPDQ",
                "QRLIFAGKQL",
            ],
        ),
        (2, "DATA", ["SEQUENCE", "EDGRTLSDYN", "IQKESTLHLV", "LRLRGG"]),
        (
            1,
            "VARS",
            [
                "RESID_I",
                "RESNAME_I",
                "ATOMNAME_I",
                "RESID_J",
                "RESNAME_J",
                "ATOMNAME_J",
                "D",
                "DD",
                "W",
            ],
        ),
        (
            1,
            "FORMAT",
            ["%5d", "%6s", "%6s", "%5d", "%6s", "%6s", "%9.3f", "%9.3f", "%.2f"],
        ),
        (1, "__VALUES__", [2, "GLN", "N", 2, "GLN", "HN", -15.524, 1.0, 1.0]),
        (2, "__VALUES__", [3, "ILE", "N", 3, "ILE", "HN", 10.521, 1.0, 1.0]),
        (3, "__VALUES__", [4, "PHE", "N", 4, "PHE", "HN", 9.648, 1.0, 1.0]),
        (4, "__VALUES__", [5, "VAL", "N", 5, "VAL", "HN", 6.082, 1.0, 1.0]),
        (5, "__VALUES__", [1, "MET", "C", 2, "GLN", "HN", 3.993, 0.333, 3.0]),
        (6, "__VALUES__", [2, "GLN", "C", 3, "ILE", "HN", -5.646, 0.333, 3.0]),
        (7, "__VALUES__", [3, "ILE", "C", 4, "PHE", "HN", 1.041, 0.333, 3.0]),
        (8, "__VALUES__", [4, "PHE", "C", 5, "VAL", "HN", 0.835, 0.333, 3.0]),
    ]

    records = []
    for (index, type, values), line_info in zip(base_records, line_info_items):
        records.append(
            DbRecord(index=index, type=type, values=values, line_info=line_info)
        )

    expected = DbFile(name="unknown", records=records)
    assert result == expected


def test_too_many_vars():
    TOO_MANY_VARS = """\
    DATA SEQUENCE MQIFVKTLTG KTITLEVEPS DTIENVKAKI QDKEGIPPDQ QRLIFAGKQL
    DATA SEQUENCE EDGRTLSDYN IQKESTLHLV LRLRGG

    VARS   RESID_I RESNAME_I ATOMNAME_I RESID_J RESNAME_J ATOMNAME_J D      DD    W
    VARS   RESID_I RESNAME_I ATOMNAME_I RESID_J RESNAME_J ATOMNAME_J D      DD    W
    """
    TOO_MANY_VARS = dedent(TOO_MANY_VARS)

    gdb_stream = io.StringIO(TOO_MANY_VARS)

    msg = """\
              bad NMRPipe db file, multiple VARS statements found
              file: unknown
              line no: 5
              line: VARS   RESID_I RESNAME_I ATOMNAME_I RESID_J RESNAME_J ATOMNAME_J D      DD    W"""
    msg = dedent(msg)

    with pytest.raises(MultipleVars, match=msg):
        read_db_file_records(gdb_stream)


def test_format_no_vars():
    NO_VARS = """\
    FORMAT %5d     %6s       %6s        %5d     %6s       %6s    %9.3f   %9.3f %.2f

        2    GLN      N      2    GLN     HN     -15.524     1.000 1.00
    """
    NO_VARS = dedent(NO_VARS)

    gdb_stream = io.StringIO(NO_VARS)

    msg = """\
              no column names defined by a VARS line when FORMAT line read
              file: unknown
              line no: 1
              line: FORMAT %5d     %6s       %6s        %5d     %6s       %6s    %9.3f   %9.3f %.2"""
    msg = dedent(msg)

    with pytest.raises(NoVarsLine, match=msg):
        read_db_file_records(gdb_stream)


def test_vars_format_mismatch():
    TOO_FEW_FORMATS = """\
    VARS   RESID_I RESNAME_I ATOMNAME_I RESID_J RESNAME_J ATOMNAME_J D      DD    W
    FORMAT %5d     %6s       %6s        %5d     %6s       %6s    %9.3f   %9.3f"""
    TOO_FEW_FORMATS = dedent(TOO_FEW_FORMATS)

    gdb_stream = io.StringIO(TOO_FEW_FORMATS)

    msgs = """\
              number of column names and formats must agree
              got 8 column names
              9 formats
              file: unknown
              line no: 2
              line: FORMAT
              """
    msgs = dedent(msgs).strip()
    msgs = msgs.split("\n")

    with pytest.raises(WrongColumnCount) as exc_info:
        read_db_file_records(gdb_stream)

    for msg in msgs:
        assert msg in exc_info.value.args[0]


def test_empty_and_comment():
    EMPTY_AND_COMMENT = """\
    #

    #
    """
    EMPTY_AND_COMMENT = dedent(EMPTY_AND_COMMENT)

    gdb_stream = io.StringIO(EMPTY_AND_COMMENT)
    db_records = read_db_file_records(gdb_stream)

    assert len(db_records.records) == 2

    REMARK = """REMARK"""

    gdb_stream = io.StringIO(REMARK)
    db_records = read_db_file_records(gdb_stream)

    assert len(db_records.records) == 1


def test_multiple_format_lines():
    MULTIPLE_FORMAT = """\
        VARS   RESID_I RESNAME_I ATOMNAME_I RESID_J RESNAME_J ATOMNAME_J D      DD    W
        FORMAT %5d     %6s       %6s        %5d     %6s       %6s    %9.3f   %9.3f %.2f
        FORMAT %5d     %6s       %6s        %5d     %6s       %6s    %9.3f   %9.3f %.2f
    """

    MULTIPLE_FORMAT = dedent(MULTIPLE_FORMAT)

    gdb_stream = io.StringIO(MULTIPLE_FORMAT)

    msgs = """\
            bad NMRPipe db file, multiple FORMAT statements found
            file: unknown
            line no: 3
            line: FORMAT
            """
    msgs = dedent(msgs).strip()
    msgs = msgs.split("\n")

    with pytest.raises(MultipleFormat) as exc_info:
        read_db_file_records(gdb_stream)

    for msg in msgs:
        assert msg in exc_info.value.args[0]


def test_too_few_columns():

    TOO_FEW_COLUMNS = """\
                        DATA SEQUENCE MQIFVKTLTG KTITLEVEPS DTIENVKAKI QDKEGIPPDQ QRLIFAGKQL
                        DATA SEQUENCE EDGRTLSDYN IQKESTLHLV LRLRGG

                        VARS   RESID_I RESNAME_I ATOMNAME_I RESID_J RESNAME_J ATOMNAME_J D      DD    W
                        FORMAT %5d     %6s       %6s        %5d     %6s       %6s    %9.3f   %9.3f %.2f

                            2    GLN      N      2    GLN     HN     -15.524     1.000
                        """

    TOO_FEW_COLUMNS = dedent(TOO_FEW_COLUMNS)

    gdb_stream = io.StringIO(TOO_FEW_COLUMNS)

    msgs = """\
            number fields (9) doesn't not match number of columns (10)
            expected
            RESID_I  RESNAME_I  ATOMNAME_I  RESID_J  RESNAME_J  ATOMNAME_J  D        DD     W
            int      str        str         int      str        str         float    float  float
            2        GLN        N           2        GLN        HN          -15.524  1.000  *
            file: unknown
            line no: 7
            line:     2    GLN
            """
    msgs = dedent(msgs).strip()
    msgs = msgs.split("\n")

    with pytest.raises(WrongColumnCount) as exc_info:
        read_db_file_records(gdb_stream)

    for msg in msgs:
        assert msg in exc_info.value.args[0]


def test_data_before_format():
    DATA_BEFORE_FORMAT = """\
                            2    GLN      N      2    GLN     HN     -15.524     1.000
                        """

    DATA_BEFORE_FORMAT = dedent(DATA_BEFORE_FORMAT)

    gdb_stream = io.StringIO(DATA_BEFORE_FORMAT)

    msgs = """\
            bad nmrpipe db file, data seen before VAR and FORMAT
            file: unknown
            line no: 1
            line: 2    GLN
            """
    msgs = dedent(msgs).strip()
    msgs = msgs.split("\n")

    with pytest.raises(DataBeforeFormat) as exc_info:
        read_db_file_records(gdb_stream)

    for msg in msgs:
        assert msg in exc_info.value.args[0]


def test_bad_data_format():
    BAD_DATA_FORMAT = """\
                        VARS   RESID_I RESNAME_I ATOMNAME_I RESID_J RESNAME_J ATOMNAME_J D      DD    W
                        FORMAT %5d     %6s       %6s        %5d     %6s       %6s    %9.3f   %9.3f %.2f

                            2    GLN      N      GLN     2      HN     -15.524     1.000  1.00
                        """

    BAD_DATA_FORMAT = dedent(BAD_DATA_FORMAT)

    gdb_stream = io.StringIO(BAD_DATA_FORMAT)

    msgs = """\
            Couldn't convert GLN to type int
            file: unknown
            line no: 4
            column: 4
            line:     2    GLN      N      GLN     2      HN     -15.524     1.000  1.00
                                           ^
            """
    msgs = dedent(msgs).strip()
    msgs = msgs.split("\n")

    with pytest.raises(BadFieldFormat) as exc_info:
        read_db_file_records(gdb_stream)

    for msg in msgs:
        assert msg in exc_info.value.args[0]


def test_bad_format():
    BAD_FIELD_FORMAT = """\
                        VARS   RESID_I RESNAME_I ATOMNAME_I RESID_J RESNAME_J ATOMNAME_J D      DD    W
                        FORMAT %5g     %6s       %6s        %5d     %6s       %6s    %9.3f   %9.3f %.2f

                            2    GLN      N      GLN     2      HN     -15.524     1.000  1.00
                        """
    BAD_FIELD_FORMAT = dedent(BAD_FIELD_FORMAT)

    gdb_stream = io.StringIO(BAD_FIELD_FORMAT)

    with pytest.raises(BadFieldFormat) as exc_info:
        read_db_file_records(gdb_stream)

    msgs = """\
          unexpected format g at index 1
          expected formats are
          s, d, e, f
          string
          integer
          scientific(float)
          float
          file: unknown
          line no: 2
          ^
          """
    msgs = dedent(msgs)
    msgs = msgs.split("\n")

    for msg in msgs:
        assert msg.strip() in exc_info.value.args[0]


def test_columns():
    TEST_DATA = """\
                        DATA SEQUENCE MQIFVKTLTG KTITLEVEPS DTIENVKAKI QDKEGIPPDQ QRLIFAGKQL
                        DATA SEQUENCE EDGRTLSDYN IQKESTLHLV LRLRGG

                        VARS   RESID_I RESNAME_I ATOMNAME_I RESID_J RESNAME_J ATOMNAME_J D      DD    W
                        FORMAT %5d     %6s       %6s        %5d     %6s       %6s    %9.3f   %9.3f %.2f

                            2    GLN      N      2    GLN     HN     -15.524     1.000 1.000
                        """

    TEST_DATA = dedent(TEST_DATA)

    gdb_stream = io.StringIO(TEST_DATA)
    gdb_records = read_db_file_records(gdb_stream)

    columns = get_gdb_columns(gdb_records)

    assert (
        columns
        == "RESID_I RESNAME_I ATOMNAME_I RESID_J RESNAME_J ATOMNAME_J D      DD    W".split()
    )


def test_column_indices():
    TEST_DATA = """\
                        DATA SEQUENCE MQIFVKTLTG KTITLEVEPS DTIENVKAKI QDKEGIPPDQ QRLIFAGKQL
                        DATA SEQUENCE EDGRTLSDYN IQKESTLHLV LRLRGG

                        VARS   RESID_I RESNAME_I ATOMNAME_I RESID_J RESNAME_J ATOMNAME_J D      DD    W
                        FORMAT %5d     %6s       %6s        %5d     %6s       %6s    %9.3f   %9.3f %.2f

                            2    GLN      N      2    GLN     HN     -15.524     1.000 1.000
                        """

    TEST_DATA = dedent(TEST_DATA)

    gdb_stream = io.StringIO(TEST_DATA)
    gdb_records = read_db_file_records(gdb_stream)

    columns = get_column_indices(gdb_records)

    column_names = "RESID_I RESNAME_I ATOMNAME_I RESID_J RESNAME_J ATOMNAME_J D      DD    W".split()
    assert columns == {
        column_name: index for index, column_name in enumerate(column_names)
    }


def test_column_data():
    TEST_DATA = """\
                        DATA SEQUENCE MQIFVKTLTG KTITLEVEPS DTIENVKAKI QDKEGIPPDQ QRLIFAGKQL
                        DATA SEQUENCE EDGRTLSDYN IQKESTLHLV LRLRGG

                        VARS   RESID_I RESNAME_I ATOMNAME_I RESID_J RESNAME_J ATOMNAME_J D      DD    W
                        FORMAT %5d     %6s       %6s        %5d     %6s       %6s    %9.3f   %9.3f %.2f

                            2    GLN      N      2    GLN     HN     -15.524     1.000 1.000
                        """

    TEST_DATA = dedent(TEST_DATA)

    gdb_stream = io.StringIO(TEST_DATA)
    gdb_file = read_db_file_records(gdb_stream)

    column_indices = get_column_indices(gdb_file)
    print("!!", column_indices)

    data = select_records(gdb_file, VALUES)
    print("!! data", data[0])

    column_names = "RESID_I RESNAME_I ATOMNAME_I RESID_J RESNAME_J ATOMNAME_J D      DD    W".split()
    print(len(column_names), len(data[0].values))

    values = [
        data[0].values[column_indices[column_name]] for column_name in column_names
    ]

    assert values == [2, "GLN", "N", 2, "GLN", "HN", -15.524, 1.000, 1.000]

    # assert False


def test_sequence():

    SEQUENCE = f"DATA {dedent(ABC_SEQUENCE_1LET)}"

    gdb_stream = io.StringIO(SEQUENCE)

    records = read_db_file_records(gdb_stream)
    sequence = gdb_to_3let_sequence(records)

    sequence == ABC_SEQUENCE_3LET


if __name__ == "__main__":
    pytest.main([__file__, "-vv"])



import typer
from typer.testing import CliRunner

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.xplor.exporters.rdcs import rdcs

runner = CliRunner()
app = typer.Typer()
app.command()(rdcs)

EXPECTED = """\
! restraints from frame nef_rdc_restraint_list_rdcs in nef entry xplor

assign  (  resn   ANI   and  name   OO  )
        (  resn   ANI   and  name   Z   )
        (  resn   ANI   and  name   X   )
        (  resn   ANI   and  name   Y   )
        (  segid  AAAA  and  resid  1   and  resname  ALA  and  name  HA    )
        (  segid  AAAA  and  resid  1   and  resname  ALA  and name   HN    )    1.300   1.000
assign  (  resn   ANI   and  name   OO  )
        (  resn   ANI   and  name   Z   )
        (  resn   ANI   and  name   X   )
        (  resn   ANI   and  name   Y   )
        (  segid  AAAA  and  resid  1   and  resname  ALA  and  name  HB    )
        (  segid  AAAA  and  resid  1   and  resname  ALA  and  name  HN    )    4.600   1.000
assign  (  resn   ANI   and  name   OO  )
        (  resn   ANI   and  name   Z   )
        (  resn   ANI   and  name   X   )
        (  resn   ANI   and  name   Y   )
        (  segid  AAAA  and  resid  2   and  resname  ALA  and  name  HG3#  )
        (  segid  AAAA  and  resid  2   and  resname  ALA  and  name  HN    )    2.400   1.000
assign  (  resn   ANI   and  name   OO  )
        (  resn   ANI   and  name   Z   )
        (  resn   ANI   and  name   X   )
        (  resn   ANI   and  name   Y   )
        (  segid  AAAA  and  resid  3   and  resname  ALA  and  name  HA    )
        (  segid  AAAA  and  resid  3   and  resname  ALA  and  name  HN    )    6.700   1.000

"""


# noinspection PyUnusedLocal
def test_3ab(clear_cache):

    input = open(path_in_test_data(__file__, "3a_ab_rdcs.neff")).read()
    result = run_and_report(app, [], input=input)

    assert_lines_match(EXPECTED, result.stdout)



import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.tools.chains.list import list as list_app

app = typer.Typer()
app.command()(list_app)


# noinspection PyUnusedLocal
def test_frame_basic(clear_cache):

    path = path_in_test_data(__file__, "multi_chain.nef")

    result = run_and_report(app, [], input=open(path).read())

    EXPECTED = "A B C"

    assert_lines_match(EXPECTED, result.stdout)



import contextlib
import inspect
import io
import os
import sys
import traceback
from argparse import Namespace
from enum import auto
from pathlib import Path
from textwrap import dedent
from typing import Any, Dict, Iterator, List, Optional, TextIO, TypeVar, Union

import click
from cacheable_iter import iter_cache
from pynmrstar import Entry, Loop, Saveframe
from strenum import StrEnum
from tabulate import tabulate

from nef_pipelines.lib.constants import (
    EXIT_ERROR,
    NEF_META_DATA,
    NEF_PIPELINES,
    NEF_UNKNOWN,
)
from nef_pipelines.lib.header_lib import (
    create_header_frame,
    get_creation_time,
    get_uuid,
)
from nef_pipelines.lib.structures import LineInfo

STDIN = Path("-")
STDOUT = Path("-")


def _get_loop_by_category_or_none(frame: Saveframe, category: str) -> Loop:

    result = None
    if f"_{category}" in frame.loop_dict.keys():
        result = frame.get_loop(category)

    return result


def loop_add_data_tag_dict(loop: Loop, data: Dict[str, object]) -> None:
    tagged_data = []
    for category_tag in loop.get_tag_names():
        _, tag = category_tag.split(".")
        if tag in data:
            tagged_data.append(data[tag])
        else:
            tagged_data.append(NEF_UNKNOWN)

    loop.add_data(tagged_data)


# TODO: there is no space for scrript arguments!
def fixup_metadata(entry: Entry, name: str, version: str, script: str):
    """
    add a new entry to a nef metadata frame

    Args:
        entry (Entry): the target entry
        name (str): the name of the program used
        version (str): the vserion of the program used
        script (str): the script used

    Returns:
        Entry: the modified entry
    """

    if entry is not None and NEF_META_DATA in entry.category_list:
        meta_frame = entry[NEF_META_DATA]

        last_program = meta_frame.get_tag("program_name")[0]
        last_program_version = meta_frame.get_tag("program_version")[0]
        last_script_name = meta_frame.get_tag("script_name")[0]

        meta_frame.add_tag("program_name", name, update=True)
        meta_frame.add_tag("program_version", version, update=True)
        meta_frame.add_tag("script_name", script, update=True)
        creation_time = get_creation_time()
        meta_frame.add_tag("creation_time", creation_time, update=True)
        uuid = get_uuid(NEF_PIPELINES, creation_time)
        meta_frame.add_tag("uuid", uuid, update=True)

        run_history_loop = _get_loop_by_category_or_none(meta_frame, "nef_run_history")
        if run_history_loop is not None:
            if run_history_loop.get_tag("run_number"):
                run_number_tags = run_history_loop.get_tag("run_number")
                run_numbers = [int(run_number) for run_number in run_number_tags]
                last_run_number = max(run_numbers)
                next_run_number = last_run_number + 1
            else:
                next_run_number = 1

            data = {
                "run_number": next_run_number,
                "program_name": last_program,
                "program_version": last_program_version,
                "script_name": last_script_name,
            }
            loop_add_data_tag_dict(run_history_loop, data)
        else:
            run_history_loop = Loop.from_scratch("nef_run_history")
            run_history_loop.add_tag(
                ["run_number", "program_name", "program_version", "script_name"]
            )
            run_history_loop.add_data(["."] * 4)
    else:
        header = create_header_frame(NEF_PIPELINES, get_version(), script)
        entry.add_saveframe(header)


class StringIteratorIO(io.TextIOBase):

    # https://stackoverflow.com/questions/12593576/adapt-an-iterator-to-behave-like-a-file-like-object-in-python
    # second answer with more votes!

    def __init__(self, iter):
        self._iter = iter
        self._left = ""

    def readable(self):
        return True

    def _read1(self, n=None):
        while not self._left:
            try:
                self._left = next(self._iter)
            except StopIteration:
                break
        ret = self._left[:n]
        self._left = self._left[len(ret) :]
        return ret

    def read(self, n=None):
        line = []
        if n is None or n < 0:
            while True:
                m = self._read1()
                if not m:
                    break
                line.append(m)
        else:
            while n > 0:
                m = self._read1(n)
                if not m:
                    break
                n -= len(m)
                line.append(m)
        return "".join(line)

    def readline(self):
        line = []
        while True:
            i = self._left.find("\n")
            if i == -1:
                line.append(self._left)
                try:
                    self._left = next(self._iter)
                except StopIteration:
                    self._left = ""
                    break
            else:
                line.append(self._left[: i + 1])
                self._left = self._left[i + 1 :]
                break
        return "".join(line)


# https://stackoverflow.com/questions/6657820/how-to-convert-an-iterable-to-a-stream/20260030#20260030
def iterable_to_stream(iterable, buffer_size=io.DEFAULT_BUFFER_SIZE):
    """
    Lets you use an iterable (e.g. a generator) that yields bytestrings as a read-only
    input stream.

    The stream implements Python 3's newer I/O API (available in Python 2's io module).
    For efficiency, the stream is buffered.
    """

    return StringIteratorIO(iterable)


def running_in_pycharm():
    return "PYCHARM_HOSTED" in os.environ


def get_text_from_file_or_exit(file_name: Path) -> str:
    """
    get text from stdin or from a file argument

    Args:
        file_name (str): name of the file Path('-') indicates stdin

    Returns:
        str: text from file

    """

    result = None

    if file_name == Path("-") or not file_name:
        if sys.stdin.isatty():
            exit_error("attempting to read from a terminal")
        result = sys.stdin.read()
    else:
        try:
            result = open(file_name).read()
        except IOError as e:
            exit_error(f"couldn't read from {file_name} because of an error", e)

    return result


def get_pipe_file(args: Namespace) -> Optional[TextIO]:
    """
    get an input on stdin or from an argument called pipe

    Args:
        args (Namespace): command line arguments

    Returns:
        TextIO: an input stream or None

    """

    result = None
    if "pipe" in args and args.pipe:
        result = iter(cached_file_stream(args.pipe))
    # pycharm doesn't treat stdstreams correcly and hangs
    elif not sys.stdin.isatty() and not running_in_pycharm():
        result = iter(sys.stdin.read().split("\n"))

    return StringIteratorIO(result) if result else None


def get_pipe_file_text(args: Namespace) -> Optional[str]:
    """
    get an input on stdin or from an argument called pipe

    Args:
        args (Namespace): command line arguments

    Returns:
        TextIO: an input stream or None

    """

    result = None
    if "pipe" in args and args.pipe:
        result = open(args.pipe, "r").read()
    # pycharm doesn't treat stdstreams correcly and hangs
    elif not sys.stdin.isatty() and not running_in_pycharm():
        result = sys.stdin.read()

    return result if result else None


def get_pipe_file_text_or_exit(args: Namespace) -> Optional[TextIO]:

    try:
        result = get_pipe_file_text(args)
    except Exception as e:
        exit_error("couldn't read from stdin or -pipe file", e)

    if result is None:
        exit_error("couldn't read from stdin and no -pipe in command line arguments")

    return result


def get_pipe_file_or_exit(args: Namespace) -> Optional[TextIO]:

    try:
        result = get_pipe_file(args)
    except Exception as e:
        exit_error("couldn't read from stdin or -pipe file", e)

    if isinstance(result, StringIteratorIO):

        result = "\n".join(list(result))

        if len(result) == 0:
            result = None
        else:
            result = io.StringIO(result)

    if result is None:
        exit_error("couldn't read from stdin and no -pipe in command line arguments")

    return result


@iter_cache
def cached_stdin():
    # TODO: this doesn't work!'
    return sys.stdin


class cached_file_stream:
    def __init__(self, file_name):
        self._file_name = file_name

    def __enter__(self):
        return _cached_file_stream(self._file_name)

    def __exit__(self, *args):
        pass

    def __iter__(self):
        return _cached_file_stream(self._file_name)


@iter_cache
def _cached_file_stream(file_name):
    try:
        with open(file_name, "r") as lines:
            result = lines.readlines()
    except IOError as e:
        exit_error(f"couldn't open stream {file_name} because {e}")
    return result


def script_name(file: str) -> Path:
    """
    get the name of the script

    Args:
        file (str): the name of the file

    Returns:
        Path: path to the file
    """
    try:
        current_context = click.get_current_context()
    except RuntimeError:

        current_context = Namespace(command_path="unknown unknown")

    in_pytest = "pytest" in sys.modules
    if in_pytest:
        command = current_context.command_path.split()
    else:
        command = current_context.command_path.split()[1:]

    command = f"{'/'.join(command)}.py"

    return command


def read_from_file_or_exit(file_name: Path, target: str) -> str:
    if file_name == STDIN and sys.stdin.isatty():
        exit_error(
            f"trying to open stdin to read {target}, but stdin is not a stream [did you forget to add a nef file to "
            "your pipeline?]"
        )
    if running_in_pycharm():
        exit_error(
            "reading from stdin doesn't work in pycharm debug environment as there is no shell..."
        )

    display_file_name = get_display_file_name(file_name)

    fh = sys.stdin

    if not file_name == STDIN:
        try:
            fh = open(file_name)
        except Exception as e:
            msg = f"couldn't open {display_file_name} to read {target}"
            exit_error(msg, e)

    try:
        text = fh.read()
    except Exception as e:
        msg = f"couldn't read {target} from {display_file_name}"
        exit_error(msg, e)

    return text


def _script_to_command(script: str) -> str:
    """
    turns a script path into the equivalent nef pipelines command
    :param script: the script path
    :return: the command name
    """
    script_path = script.split("/")
    script_path[-1] = script_path[-1][: -len(".py")]
    return " ".join(script_path)


def exit_error(msg, exception=None):
    """
    print an error message and exit error

    Args:
        msg: the message
    """

    if exception is not None:
        exc_info = sys.exc_info()
        traceback.print_exception(*exc_info, file=sys.stderr)
        print(file=sys.stderr)

    msg = dedent(msg).split("\n")

    print(file=sys.stderr)

    script = script_name(__file__)

    command = _script_to_command(script)

    print(f"ERROR [in: {command}]: {msg[0]}", file=sys.stderr)

    for line in msg[1:]:
        print(f"       {line}", file=sys.stderr)
    print("exiting...", file=sys.stderr)
    sys.exit(EXIT_ERROR)


def process_stream_and_add_frames(
    frames: List[Saveframe], input_args: Namespace
) -> Entry:
    """
    take a set of save frames and either add them to a stream from stdin / a pipe-file or create
    a new stream with a proper NEF metadata header

    Args:
        frames: a set of save frames
        input_args: command line arguments for an entry_name and pipe file source

    Returns:
        a new entry containing the frames
    """

    try:
        stream = get_pipe_file(input_args)
    except Exception as e:
        exit_error(f"failed to load pipe file because {e}")

    if stream is not None:
        lines = "\n".join(stream.readlines())
    else:
        lines = ""

    if len(lines.strip()) == 0:
        stream = None

    new_entry = (
        Entry.from_string(lines)
        if stream
        else Entry.from_scratch(input_args.entry_name)
    )

    fixup_metadata(new_entry, NEF_PIPELINES, get_version(), script_name(__file__))

    # TODO: check if frame exists
    for frame in frames:
        new_entry.add_saveframe(frame)

    return new_entry


def is_int(value: str) -> bool:
    """
    Check if a string is an integer
    Args:
        value (str): the putative integer

    Returns bool:
        true is if its an integer

    """
    result = False
    try:
        int(value)
        result = True
    except (ValueError, TypeError):
        pass

    return result


def is_float(value: str) -> bool:
    """
    Check if a string is an float
    Args:
        value (str): the putative float

    Returns bool:
        true is if its an integer

    """
    result = False
    try:
        float(value)
        result = True
    except (ValueError, TypeError):
        pass

    return result


T = TypeVar("T")


# https://stackoverflow.com/questions/312443/how-do-i-split-a-list-into-equally-sized-chunks
def chunks(input: Iterator[T], n: int) -> Iterator[List[T]]:
    """Yield successive n-sized chunks from lst.
       lst: the list to chunk
       n: the chunk size

    Returns:
        an iterator of chunks from lst of length T

    """
    lst = list(input)
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def _build_int_float_error_message(message, line_info, field, format):
    messages = []
    if message:
        messages.append(message)
    if line_info:
        message = f"""
                    couldn't convert {line_info.line} to format

                    at {line_info.line_no + 1} in f{line_info.file_name}

                    line value was

                    {line_info.line}

                    field was

                    {field}
                """
        messages.append(message)
    message = "\n\n".join(messages)
    return message


def read_float_or_exit(
    string: str, line_info: LineInfo = None, field="unknown", message=None
) -> float:
    """
    convert a string to an float or exit with error including line information

    Args:
        string: string to parse
        line_info: the line the string came from
        field: the name of the field defininbg the string (ignored if there is no line info)
        message: any other messages to show

    Returns:
        a float
    """

    format = "float"
    try:
        result = float(string)
    except Exception:
        message = _build_int_float_error_message(message, line_info, field, format)

        exit_error(message)

    return result


def read_integer_or_exit(
    string: str, line_info: LineInfo = None, field="unknown", message=None
) -> int:
    """
    convert a string to an int or exit with error including line information

    Args:
        string: string to parse
        line_info: the line the string came from
        field: the name of the field defining the string (ignored if there is no line info)
        message: any other messages to show

    Returns:
        an integer
    """

    format = "integer"
    try:
        result = int(string)
    except Exception:

        message = _build_int_float_error_message(message, line_info, field, format)

        exit_error(message)

    return result


# TODO test this!
def parse_comma_separated_options(
    lists: Union[List[List[str]], List[str], str]
) -> List[str]:
    """
    Take a mixed list of strings or strings that can be parsed as a comma separated list of strings
    and make a list of all the items

    e.g.

    test = [['1','2','3'], 'a,b,c', ['4','5']]

    gives

    ['1','2','3','a','b','c','4','5']

    :param lists: a mixture of  lists of string or comma separated as a at sting
    :return: list of items
    """
    result = []
    for item in lists:
        if isinstance(item, str):
            item = item.strip(",")
            result.extend(item.split(","))
        else:
            result.append(item)

    return result


def end_with_ordinal(n):
    return str(n) + {1: "st", 2: "nd", 3: "rd"}.get(
        4 if 10 <= n % 100 < 20 else n % 10, "th"
    )


def get_display_file_name(file_name: Path) -> str:
    return "stdin" if file_name == STDIN else file_name


class StdStream(StrEnum):
    STDIN = auto()
    STDOUT = auto()
    STDERR = auto()


STREAM_MODES = {StdStream.STDIN: "r", StdStream.STDOUT: "w", StdStream.STDERR: "w"}


@contextlib.contextmanager
def smart_open(filename=None, stream_type=StdStream.STDOUT):
    f"""
    context manager to open a file or use stdout depending on the filename, stdin is defined as '-'
    by corollary a std stream won't be closed on exit only file system streams
    :param filename: the filename to use '-' indicates a standar stream selected by stream_type
    :param stream_type: how to treat the stream choices are {', '.join(StdStream.__members__)}

    """

    mode = STREAM_MODES[stream_type]
    if filename and str(filename) != "-":
        fh = open(filename, mode)
    else:
        fh = sys.stdout

    try:
        yield fh
    finally:
        if fh not in (sys.stdout, sys.stdin, sys.stderr):
            fh.close()


class DebugStdout:
    """
    class to intercept stdout and print debugging info of where strings ae printed from

    usage:

        from contextlib import redirect_stdout
        with redirect_stdout(DebugStdout):
            ... # your code here
    """

    def __init__(self, in_old_stdout):
        self.old_stdout = in_old_stdout

    def write(self, msg):
        self.old_stdout.write(msg)
        from inspect import currentframe, getframeinfo

        frameinfo = getframeinfo(currentframe().f_back)

        print(f" <-[{frameinfo.filename}, {frameinfo.lineno}]", file=sys.stderr)


def flatten(in_list: Union[List[Any], List[Union[List[Any], Any]]]) -> List[Any]:
    """
    flatten a list of lists into a single list
    :param in_list: a putative lists of lists
    :return: a list of anything but lists
    """
    out = []
    for elem in in_list:
        if not isinstance(elem, str):
            out.extend(elem)
        else:
            out.append(elem)
    return out


def get_version() -> str:
    """
        get the current version of nef pipelines
    :return: a version string
    """

    file_path = Path(__file__)
    root_path = file_path.parent.parent
    version_path = root_path / "VERSION"

    with open(version_path) as file_h:
        version = file_h.read().strip()

    return version


class FStringTemplate:
    """class for makes an f string based template substitution is delayed
    until the instance is converted to a string
    """

    def __init__(self, template):
        self.template = template

    def __str__(self):
        variables = inspect.currentframe().f_back.f_globals.copy()
        variables.update(inspect.currentframe().f_back.f_locals)
        return self.template.format(**variables)


def unused_to_none(value: str) -> str:
    """
    If the value is a NEF UNUSED [.] replace it with None
    :param value: value which might be UNUSED
    :return: value or None if the value was UNUSED
    """
    from nef_pipelines.lib.nef_lib import (
        UNUSED,  # to avoid circular import, move to constants?
    )

    if value == UNUSED:
        value = None
    return value


def unused_to_empty_string(value):
    """
    If the value is a NEF UNUSED [.] replace it with the empty string
    :param value: value which might be UNUSED
    :return: value or "" if the value was UNUSED
    """
    from nef_pipelines.lib.nef_lib import (
        UNUSED,  # to avoid circular import , move to constants?
    )

    if value == UNUSED:
        value = ""
    return value


def _row_to_table(rows: Dict[str, str], headers=("tag", "value")) -> List[List[str]]:
    """
    convert a dictionary into a table with tabulate with two columns containing
    keys and values
    :param rows: a dictionary of key value pairs [tags and values...]
    :return: a formatted tabulate table as a string
    """
    values = [
        headers,
    ]
    for key, value in rows.items():
        values.append([key, value])
    value_string = tabulate(values, headers="firstrow")
    return value_string



# TODO: add reading of sequence from input stream
# TODO: xplor names -> iupac and deal with %% and ## properly
# TODO: add common experiment types
# TODO: guess axis codes
# TODO: add a chemical shift list reference
# TODO: _nef_nmr_spectrum: value_first_point, folding, absolute_peak_positions, is_acquisition
# TODO: cleanup
# TODO: add function
# TODO: remove ics
# TODO: multiple assignments per peak... howto in nef
# TODO: add libs pipeline
# TODO axis codes need to be guessed
# TODO doesn't check if peaks are deleted (should have an option to read all peaks?)

import itertools
from collections import OrderedDict
from pathlib import Path
from typing import Dict, List, Tuple

import typer
from ordered_set import OrderedSet
from pynmrstar import Entry, Loop, Saveframe

from nef_pipelines.lib import constants
from nef_pipelines.lib.constants import NEF_UNKNOWN
from nef_pipelines.lib.structures import (
    AtomLabel,
    PeakAxis,
    PeakList,
    PeakListData,
    PeakValues,
    SequenceResidue,
)
from nef_pipelines.lib.typer_utils import get_args
from nef_pipelines.lib.util import (
    exit_error,
    get_pipe_file,
    process_stream_and_add_frames,
)
from nef_pipelines.transcoders.nmrview import import_app

from ..nmrview_lib import parse_float_list, parse_tcl
from .sequence import read_sequence

app = typer.Typer()


# noinspection PyUnusedLocal
@import_app.command(no_args_is_help=True)
def peaks(
    entry_name: str = typer.Option(
        "nmrview",
        "-n",
        "--name",
        help="a name for a shift frame, additional calls add more names",
    ),
    chain_code: str = typer.Option(
        "A", "--chain", help="chain code", metavar="<chain-code>"
    ),
    sequence: str = typer.Option(
        None,
        "-s",
        "--sequence",
        metavar="<nmrview>.seq)",
        help="seq file for the chain <seq-file>.seq",
    ),
    axis_codes: str = typer.Option(
        "1H.15N",
        "-a",
        "--axis",
        metavar="<axis-codes>",
        help="a list of axis codes joined by dots",
    ),
    file_names: List[Path] = typer.Argument(
        ..., help="input peak files", metavar="<peak-file.xpk>"
    ),
):
    """convert nmrview peak file <nmrview>.xpk files to NEF"""
    args = get_args()

    raw_sequence = _get_sequence_or_exit(args)
    sequence = _sequence_to_residue_type_lookup(raw_sequence)

    frames = [
        read_xpk_file(args, sequence),
    ]

    entry = process_stream_and_add_frames(frames, args)
    print(entry)


def read_xpk_file(args, sequence, entry_name=None):

    with open(args.file_names[0], "r") as lines:
        peaks_list = read_raw_peaks(lines, args.chain_code, sequence)

    if not entry_name:
        entry_name = make_peak_list_entry_name(peaks_list)

    return create_spectrum_frame(args, entry_name, peaks_list)


def read_raw_peaks(lines, chain_code, sequence):

    header = get_header_or_exit(lines)

    header_data = read_header_data(lines, header)

    column_indices = read_peak_columns(lines, header_data)

    raw_peaks = read_peak_data(lines, header_data, column_indices, chain_code, sequence)

    return PeakList(header_data, raw_peaks)


def read_peak_data(lines, header_data, column_indices, chain_code, sequence):
    raw_peaks = []
    field = None
    axis_index = None
    for line_no, raw_line in enumerate(lines):
        if not len(raw_line.strip()):
            continue
        try:
            peak = {}
            line = parse_tcl(raw_line)
            # TODO validate and report errors
            peak_index = int(line[0])

            for axis_index, axis in enumerate(header_data.axis_labels):
                axis_values = read_axis_for_peak(
                    line, axis, column_indices, chain_code, sequence
                )

                peak[axis_index] = PeakAxis(*axis_values)

            raw_values = read_values_for_peak(line, column_indices)
            peak["values"] = PeakValues(peak_index, **raw_values)

            raw_peaks.append(peak)

        except Exception as e:
            field = str(field) if field else "unknown"
            msg = (
                f"failed to parse file a line {line_no} with input: '{raw_line.strip()}' field: {field}  axis: "
                f"  {axis_index + 1} exception: {e}"
            )
            exit_error(msg)
    return raw_peaks


def read_peak_columns(lines, header_data):
    line = next(lines)
    raw_headings = line.split()
    heading_indices = OrderedDict({"index": 0})
    for axis_index, axis in enumerate(header_data.axis_labels):
        for axis_field in list("LPWBEJU"):
            header = f"{axis}.{axis_field}"
            if header in raw_headings:
                heading_indices[header] = raw_headings.index(header) + 1
    for peak_item in ["vol", "int", "stat", "comment", "flag0"]:
        if peak_item in raw_headings:
            heading_indices[peak_item] = raw_headings.index(peak_item) + 1
    return heading_indices


def read_header_data(lines, headers):
    data_set = None
    sweep_widths = []
    spectrometer_frequencies = []
    num_axis = None
    axis_labels = None
    for header_no, header_type in enumerate(headers):
        line = next(lines)
        if header_type == "label":
            axis_labels = line.strip().split()
            num_axis = len(axis_labels)
        elif header_type == "dataset":
            data_set = line.strip()
        elif header_type == "sw":
            line_no = header_no + 2
            sweep_widths = parse_float_list(line, line_no)
            check_num_fields(sweep_widths, num_axis, "sweep widths", line, line_no)
        elif header_type == "sf":
            line_no = header_no + 2
            spectrometer_frequencies = parse_float_list(line, line_no)
            check_num_fields(
                spectrometer_frequencies,
                num_axis,
                "spectrometer frequencies",
                line,
                line_no,
            )

    # sweep widths are in ppm for nef!
    sweep_widths = [float(sweep_width) for sweep_width in sweep_widths]
    spectrometer_frequencies = [
        float(spectrometer_frequency)
        for spectrometer_frequency in spectrometer_frequencies
    ]

    for i, (sweep_width, spectrometer_frequency) in enumerate(
        zip(sweep_widths, spectrometer_frequencies)
    ):
        sweep_widths[i] = f"{sweep_width / spectrometer_frequency:.4f}"

    # TODO: peaks shifts, spectrometer frequencies how many decimal points
    spectrometer_frequencies = [
        f"{spectrometer_frequency:4}"
        for spectrometer_frequency in spectrometer_frequencies
    ]

    peak_list_data = PeakListData(
        num_axis, axis_labels, data_set, sweep_widths, spectrometer_frequencies
    )
    return peak_list_data


def get_header_or_exit(lines):
    header_items = ["label", "dataset", "sw", "sf"]

    line = next(lines)

    headers = []
    if line:
        headers = line.strip().split()

    if len(headers) != 4:
        msg = f"""this doesn't look like an nmrview xpk file,
                  i expected a header containing 4 items on the first line: {','.join(header_items)}
                  i got {line} at line 1"""
        exit_error(msg)

    for name in header_items:
        if name not in headers:
            msg = f"""this doesn't look like an nmrview xpk file,
                       i expected a header containing the values: {', '.join(header_items)}
                       i got '{line}' at line 1"""
            exit_error(msg)

    return headers


def read_axis_for_peak(line, axis, heading_indices, chain_code, sequence):
    axis_values = []
    for axis_field in list("LPWBEJU"):
        header = f"{axis}.{axis_field}"
        field_index = heading_indices[header]
        value = line[field_index]

        if axis_field == "L":
            label = value[0] if value else "?"
            if label == "?":
                residue_number = None
                atom_name = ""
            else:
                residue_number, atom_name = label.split(".")
                residue_number = int(residue_number)

            if residue_number:
                residue_type = sequence.setdefault((chain_code, residue_number), None)
            else:
                residue_type = ""

            if residue_number:
                atom = AtomLabel(
                    SequenceResidue(chain_code, residue_number, residue_type),
                    atom_name.upper(),
                )
            else:
                atom = AtomLabel(SequenceResidue("", None, ""), atom_name.upper())
            axis_values.append(atom)

        elif axis_field == "P":
            shift = float(value)
            axis_values.append(shift)
        elif axis_field in "WJU":
            pass
        elif axis_field == "E":
            merit = value
            axis_values.append(merit)
    return axis_values


def read_values_for_peak(line, heading_indices):
    peak_values = {}
    for value_field in ["vol", "int", "stat", "comment", "flag0"]:
        field_index = heading_indices[value_field]
        value = line[field_index]

        if value_field == "vol":
            peak_values["volume"] = float(value)
        elif value_field == "int":
            peak_values["height"] = float(value)
        elif value_field == "stat":
            peak_values["deleted"] = int(value) < 0
        elif value_field == "comment":
            comment = value[0].strip("'") if value else ""
            peak_values["comment"] = comment
        elif value_field == "flag0":
            pass

    return peak_values


def check_num_fields(fields, number, field_type, line, line_no):
    if len(fields) != number:
        msg = f"Expected {number} {field_type} got {len(fields)} for line: {line} at line {line_no}"
        exit_error(msg)


def _sequence_to_residue_type_lookup(
    sequence: List[SequenceResidue],
) -> Dict[Tuple[str, int], str]:
    result: Dict[Tuple[str, int], str] = {}
    for residue in sequence:
        result[residue.chain_code, residue.sequence_code] = residue.residue_name
    return result


def _get_isotope_code_or_exit(axis, axis_codes):
    if axis >= len(axis_codes):
        msg = f"can't find isotope code for axis {axis + 1} got axis codes {','.join(axis_codes)}"
        exit_error(msg)
    axis_code = axis_codes[axis]
    return axis_code


def sequence_from_frames(frames: Saveframe):

    residues = OrderedSet()
    for frame in frames:
        for loop in frame:
            chain_code_index = loop.tag_index("chain_code")
            sequence_code_index = loop.tag_index("sequence_code")
            residue_name_index = loop.tag_index("residue_name")

            for line in loop:
                chain_code = line[chain_code_index]
                sequence_code = line[sequence_code_index]
                residue_name = line[residue_name_index]
                residue = SequenceResidue(chain_code, sequence_code, residue_name)
                residues.append(residue)

    return list(residues)


def _get_sequence_or_exit(args):
    sequence_file = None
    if "sequence" in args:
        sequence_file = args.sequence

    if not sequence_file:
        try:
            stream = get_pipe_file(args)
            entry = Entry.from_file(stream)
            frames = entry.get_saveframes_by_category("nef_molecular_system")
            sequence = sequence_from_frames(frames)

        except Exception as e:
            exit_error(f"failed to read sequence from input stream because {e}", e)

    else:
        with open(sequence_file, "r") as lines:
            sequence = read_sequence(lines, chain_code=args.chain_code)
    return sequence


def create_spectrum_frame(args, entry_name, peaks_list):

    category = "nef_nmr_spectrum"
    frame_code = f"{category}_{entry_name}"
    frame = Saveframe.from_scratch(frame_code, category)

    frame.add_tag("sf_category", category)
    frame.add_tag("sf_framecode", frame_code)
    frame.add_tag("num_dimensions", peaks_list.peak_list_data.num_axis)
    frame.add_tag("chemical_shift_list", constants.NEF_UNKNOWN)
    loop = Loop.from_scratch("nef_spectrum_dimension")
    frame.add_loop(loop)
    list_tags = (
        "dimension_id",
        "axis_unit",
        "axis_code",
        "spectrometer_frequency",
        "spectral_width",
        "value_first_point",
        "folding",
        "absolute_peak_positions",
        "is_acquisition",
    )
    loop.add_tag(list_tags)
    list_data = peaks_list.peak_list_data
    for i in range(list_data.num_axis):
        for tag in list_tags:
            if tag == "dimension_id":
                loop.add_data_by_tag(tag, i + 1)
            elif tag == "axis_unit":
                loop.add_data_by_tag(tag, "ppm")
            elif tag == "axis_code":
                axis_codes = args.axis_codes.split(".")
                loop.add_data_by_tag(tag, _get_isotope_code_or_exit(i, axis_codes))
            elif tag == "spectrometer_frequency":
                loop.add_data_by_tag(tag, list_data.spectrometer_frequencies[i])
            elif tag == "spectral_width":
                if list_data.sweep_widths:
                    loop.add_data_by_tag(tag, list_data.sweep_widths[i])
                else:
                    loop.add_data_by_tag(tag, NEF_UNKNOWN)
            elif tag == "folding":
                loop.add_data_by_tag(tag, "circular")
            elif tag == "absolute_peak_positions":
                loop.add_data_by_tag(tag, "true")
            else:
                loop.add_data_by_tag(tag, NEF_UNKNOWN)
    loop = Loop.from_scratch("nef_spectrum_dimension_transfer")
    frame.add_loop(loop)
    transfer_dim_tags = ("dimension_1", "dimension_2", "transfer_type")
    loop.add_tag(transfer_dim_tags)
    loop = Loop.from_scratch("nef_peak")
    frame.add_loop(loop)
    peak_tags = [
        "index",
        "peak_id",
        "volume",
        "volume_uncertainty",
        "height",
        "height_uncertainty",
    ]
    position_tags = [
        (f"position_{i + 1}", f"position_uncertainty_{i + 1}")
        for i in range(list_data.num_axis)
    ]
    position_tags = itertools.chain(*position_tags)
    atom_name_tags = [
        (
            f"chain_code_{i + 1}",
            f"sequence_code_{i + 1}",
            f"residue_name_{i + 1}",
            f"atom_name_{i + 1}",
        )
        for i in range(list_data.num_axis)
    ]
    atom_name_tags = itertools.chain(*atom_name_tags)
    tags = [*peak_tags, *position_tags, *atom_name_tags]

    loop.add_tag(tags)
    for i, peak in enumerate(peaks_list.peaks):
        peak_values = peak["values"]
        if peak_values.deleted:
            continue

        for tag in tags:

            if tag == "index":
                loop.add_data_by_tag(tag, i + 1)
            elif tag == "peak_id":
                loop.add_data_by_tag(tag, peak_values.serial)
            elif tag == "volume":
                loop.add_data_by_tag(tag, peak_values.volume)
            elif tag == "height":
                loop.add_data_by_tag(tag, peak_values.height)
            elif tag.split("_")[0] == "position" and len(tag.split("_")) == 2:
                index = int(tag.split("_")[-1]) - 1
                loop.add_data_by_tag(tag, peak[index].ppm)
            elif tag.split("_")[:2] == ["chain", "code"]:
                index = int(tag.split("_")[-1]) - 1
                chain_code = peak[index].atom_labels.residue.chain_code
                chain_code = chain_code if chain_code is not None else args.chain_code
                chain_code = chain_code if chain_code else "."
                loop.add_data_by_tag(tag, chain_code)
            elif tag.split("_")[:2] == ["sequence", "code"]:
                index = int(tag.split("_")[-1]) - 1
                sequence_code = peak[index].atom_labels.residue.sequence_code
                sequence_code = sequence_code if sequence_code else "."
                loop.add_data_by_tag(tag, sequence_code)
            elif tag.split("_")[:2] == ["residue", "name"]:
                index = int(tag.split("_")[-1]) - 1

                # TODO: there could be more than 1 atom label here and this should be a list...
                residue_name = peak[index].atom_labels.residue.residue_name
                residue_name = residue_name if residue_name else "."
                loop.add_data_by_tag(tag, residue_name)
            elif tag.split("_")[:2] == ["atom", "name"]:
                index = int(tag.split("_")[-1]) - 1

                atom_name = peak[index].atom_labels.atom_name
                atom_name = atom_name if atom_name else "."
                loop.add_data_by_tag(tag, atom_name)
            else:
                loop.add_data_by_tag(tag, constants.NEF_UNKNOWN)
    return frame


# TODO: can be replaced by str.removesuffix when min python version >= 3.9
def _remove_suffix(string: str, suffix: str) -> str:

    result = string
    if string.endswith(suffix):
        result = string[: -len(suffix)]

    return result


def make_peak_list_entry_name(peaks_list):
    entry_name = peaks_list.peak_list_data.data_set.replace(" ", "_")
    entry_name = _remove_suffix(entry_name, ".nv")
    entry_name = entry_name.replace(".", "_")
    return entry_name



from nef_pipelines.lib.test_lib import assert_lines_match
from nef_pipelines.transcoders.nmrpipe.nmrpipe_lib import print_pipe_sequence


def test_print_pipe_sequence(capsys):
    TEST_DATA = ["A"] * 110

    BLOCK = "AAAAAAAAAA"
    EXPECTED = f"""\
        DATA SEQUENCE {BLOCK} {BLOCK} {BLOCK} {BLOCK} {BLOCK} {BLOCK} {BLOCK} {BLOCK} {BLOCK} {BLOCK}
        DATA SEQUENCE {BLOCK}
    """

    print_pipe_sequence(TEST_DATA)

    assert_lines_match(capsys.readouterr().out, EXPECTED)



import typer

import nef_pipelines
from nef_pipelines import nef_app

frames_app = typer.Typer()


if nef_app.app:
    nef_app.app.add_typer(
        frames_app,
        name="frames",
        help="-  carry out operations on frames in nef frames",
    )

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.tools.frames.delete  # noqa: F401
    import nef_pipelines.tools.frames.insert  # noqa: F401
    import nef_pipelines.tools.frames.list  # noqa: F401
    import nef_pipelines.tools.frames.tabulate  # noqa: F401



import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.mars.importers.shifts import shifts

app = typer.Typer()
app.command()(shifts)


def test_export_shifts_nef5_short():

    result = run_and_report(app, [path_in_test_data(__file__, "sec5_short.txt")])

    EXPECTED = """\
        save_nef_chemical_shift_list_mars
           _nef_chemical_shift_list.sf_category   nef_chemical_shift_list
           _nef_chemical_shift_list.sf_framecode  nef_chemical_shift_list_mars

           loop_
              _nef_chemical_shift.chain_code
              _nef_chemical_shift.sequence_code
              _nef_chemical_shift.residue_name
              _nef_chemical_shift.atom_name
              _nef_chemical_shift.value
              _nef_chemical_shift.value_uncertainty
              _nef_chemical_shift.element
              _nef_chemical_shift.isotope_number

             @-   @1     .   H    8.594     .   .   .
             @-   @1     .   N    133.252   .   .   .
             @-   @1     .   CA   61.159    .   .   .
             @-   @1-1   .   CA   59.291    .   .   .
             @-   @1     .   CB   33.274    .   .   .
             @-   @1-1   .   CB   37.309    .   .   .
             @-   @2     .   H    8.928     .   .   .
             @-   @2     .   N    131.397   .   .   .
             @-   @3     .   H    9.671     .   .   .
             @-   @3     .   N    130.698   .   .   .
             @-   @3     .   CA   60.735    .   .   .
             @-   @3-1   .   CA   62.418    .   .   .
             @-   @3     .   CB   39.792    .   .   .
             @-   @3-1   .   CB   69.744    .   .   .
             @-   @4     .   H    8.185     .   .   .
             @-   @4     .   N    128.894   .   .   .
             @-   @4     .   CA   53.744    .   .   .
             @-   @4-1   .   CA   54.679    .   .   .
             @-   @4     .   CB   43.207    .   .   .
             @-   @4-1   .   CB   34.05     .   .   .
             @-   @5     .   H    8.548     .   .   .
             @-   @5     .   N    128.09    .   .   .
             @-   @5     .   CA   56.292    .   .   .
             @-   @5-1   .   CA   65.574    .   .   .
             @-   @5     .   CB   32.964    .   .   .
             @-   @5-1   .   CB   69.899    .   .   .

           stop_

        save_

    """

    frame = isolate_frame(result.stdout, "nef_chemical_shift_list_mars")
    assert_lines_match(EXPECTED, frame)



import sys
from pathlib import Path
from typing import List

import typer
from pynmrstar import Entry, Saveframe
from tabulate import tabulate

from nef_pipelines.lib.nef_lib import (
    UNUSED,
    read_entry_from_stdin_or_exit,
    select_frames_by_name,
)
from nef_pipelines.lib.shift_lib import nef_frames_to_shifts
from nef_pipelines.lib.util import STDOUT, exit_error, is_int
from nef_pipelines.transcoders.xcamshift import export_app

XCAMSHIFT_ATOMS = "HN N CA CB C HA HA1 HA2".split()


# noinspection PyUnusedLocal
@export_app.command()
def shifts(
    shift_frames: List[str] = typer.Option(
        [],
        "-f",
        "--frame",
        help="selector for the shift restraint frames to use, can be called multiple times and include wild cards",
    ),
    chain: str = typer.Option(
        [],
        "-c",
        "--chain",
        help="chains to export, to add multiple chains use repeated calls  [default: 'A']",
        metavar="<CHAIN-CODE>",
    ),
    output_file: str = typer.Argument(
        None,
        help="file name to output to,  for stdout use -",
        metavar="<XPLOR-SHIFT-FILE>",
    ),
):
    """- convert nef chemical shifts to xplor (xcamshift)"""

    if len(shift_frames) == 0:
        shift_frames = ["*"]

    entry = read_entry_from_stdin_or_exit()

    output_file = (
        f"{entry.entry_id}_shifts.tab" if output_file is None else Path(output_file)
    )

    entry = pipe(entry, shift_frames, output_file)

    if not (sys.stdout.isatty() or output_file == STDOUT):
        print(entry)


def pipe(entry: Entry, shift_frames: List[Saveframe], output_file: Path) -> Entry:

    frames = entry.get_saveframes_by_category("nef_chemical_shift_list")

    frames = select_frames_by_name(frames, shift_frames)

    if len(frames) == 0:
        exit_error("no shift frames selected")

    shifts = nef_frames_to_shifts(frames)

    lines = []

    no_uncertainties = True
    for shift in shifts:
        if shift.value_uncertainty != UNUSED:
            no_uncertainties = False

    for shift in shifts:

        chain_code = shift.atom.residue.chain_code
        sequence_code = shift.atom.residue.sequence_code
        atom_name = shift.atom.atom_name
        residue_name = shift.atom.residue.residue_name

        if _is_pseudo_residue(chain_code, sequence_code):
            continue

        if atom_name not in XCAMSHIFT_ATOMS:
            continue

        value = shift.value
        value_uncertainty = shift.value_uncertainty
        chain_segid = f"segid {chain_code} and"
        assign = f"assign ( {chain_segid}  resid {sequence_code} and resn {residue_name} and name {atom_name} )"

        if no_uncertainties:
            value_uncertainty = ""
        lines.append(f"{assign} {value} {value_uncertainty}".split())

    file_h = sys.stdout if output_file == STDOUT else open(output_file, "w")

    print(tabulate(lines, tablefmt="plain", floatfmt="7.3f"), file=file_h)

    if output_file != STDOUT:
        file_h.close()

    return entry


def _is_pseudo_residue(chain_code, sequence_code):
    return (
        chain_code.startswith("@-")
        or not is_int(sequence_code)
        and sequence_code.startswith("@")
    )


def _filter_headings_by_pseudoatoms(base_headers, pseudo_residues):
    atom_names = set()
    for pseudo_residue in pseudo_residues.values():
        for pseudo_atom in pseudo_residue.values():
            offset = (
                f"-{pseudo_atom.negative_offset}"
                if pseudo_atom.negative_offset == 1
                else ""
            )
            atom_names.add(f"{pseudo_atom.atom_name.upper()}{offset}")
    headers = []
    for header in base_headers:
        if header == "" or header in atom_names:
            headers.append(header)
    return headers



import inspect
import sys
from enum import auto
from fnmatch import fnmatch
from pathlib import Path
from typing import List

import typer
from pynmrstar import Entry
from strenum import LowercaseStrEnum

from nef_pipelines.lib.util import (
    cached_stdin,
    exit_error,
    parse_comma_separated_options,
    running_in_pycharm,
)
from nef_pipelines.tools.frames import frames_app

UNDERSCORE = "_"

parser = None


class PRIORITY(LowercaseStrEnum):
    STDIN = auto()
    FILE = auto()


SELECT_HELP = """a list of frames by full name or categeory to merge from the the inserted file, can use lists for frame
               names joined with , or be called mutiple times [default is all frames] if --category is set use
               categories rather than names, if --exact isn'r fefined you can use wild cards"""


# noinspection PyUnusedLocal
@frames_app.command()
def insert(
    exact: bool = typer.Option(
        False, "-e", "--exact", help="don't treat name as a wild card"
    ),
    use_categories: bool = typer.Option(
        False,
        "-c",
        "--category",
    ),
    select: List[str] = typer.Option(None, "-s", "--select", help=SELECT_HELP),
    priority: PRIORITY = typer.Option(
        PRIORITY.FILE,
        "-p",
        "--priority",
        help=f"source of frames to prioritise, one of {PRIORITY.STDIN} and {PRIORITY.FILE}",
    ),
    nef_file_paths: List[Path] = typer.Argument(
        ..., help="nef files from which to insert frames into the current stream"
    ),
):
    """- insert frames from another nef file into the current stream"""

    select = parse_comma_separated_options(select)
    select_all = len(select) == 0

    stream_entry = _create_entry_from_stdin_or_exit(current_function())

    for nef_file_path in nef_file_paths:
        external_entry = Entry.from_file(nef_file_path.open())
        for external_frame in external_entry:

            ok_external_frame = False

            if select_all:
                ok_external_frame = True
            if use_categories and exact and external_frame.category in select:
                ok_external_frame = True
            if not use_categories and exact and external_frame.name in select:
                ok_external_frame = True
            if use_categories and not exact:
                for category in select:
                    if fnmatch(external_frame.category, f"*{category}*"):
                        ok_external_frame = True
            if not use_categories and not exact:
                for name in select:
                    if fnmatch(external_frame.name, f"*{name}*"):
                        ok_external_frame = True

            if ok_external_frame:
                if external_frame in stream_entry:
                    if priority == PRIORITY.STDIN:
                        continue
                    elif priority == PRIORITY.FILE:
                        del stream_entry[external_frame.name]
                        stream_entry.add_saveframe(external_frame)
                else:
                    stream_entry.add_saveframe(external_frame)

    print(stream_entry)


def current_function():

    return inspect.stack()[1][3]


def calling_function():

    return inspect.stack()[2][3]


def _create_entry_from_stdin_or_exit(command_name: str):

    try:

        if sys.stdin.isatty():
            exit_error(
                f"the command {command_name} reads from stdin and there is no stream..."
            )

        if running_in_pycharm():
            exit_error("you can't build read fron stdin in pycharm...")

        result = cached_stdin()

        # result is an iterable as well as an iter, but may have been read already making the iter empty?
        # hence the need to call iter?
        lines = list(iter(result))

        if len(lines) == 0:
            exit_error(
                f"the command {command_name} reads from stdin and the stream is empty..."
            )

        entry = Entry.from_string("".join(lines))

    except Exception as e:
        exit_error(f"failed to read nef entry from stdin because {e}", e)

    return entry



import typer
from typer.testing import CliRunner

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.tools.chains.renumber import renumber

runner = CliRunner()

app = typer.Typer()
app.command()(renumber)

OFFSET_CHAINS_A_10 = ["A", "10"]
OFFSET_CHAINS_A_10_B_5 = ["A", "10", "B", "5"]
OFFSET_CHAINS_MISSING_OFFSET = ["A", "10", "B"]
OFFSET_CHAINS_BAD_OFFSET = ["A", "10", "B", "B5"]
SET_CHAIN_STARTS_A_0_B_10 = ["--starts", "A", "0", "B", "10"]
SET_CHAIN_STAR_A_SHIFTS = ["--starts", "A", "10", "--frames", "nef_nmr_spectrum"]


# noinspection PyUnusedLocal
def test_renumber_basic():

    path = path_in_test_data(__file__, "tailin_seq_short.nef")

    result = run_and_report(app, [*OFFSET_CHAINS_A_10], input=open(path))

    EXPECTED = """\
        data_talin1

        save_nef_molecular_system
           _nef_molecular_system.sf_category   nef_molecular_system
           _nef_molecular_system.sf_framecode  nef_molecular_system

           loop_
              _nef_sequence.index
              _nef_sequence.chain_code
              _nef_sequence.sequence_code
              _nef_sequence.residue_name
              _nef_sequence.linking
              _nef_sequence.residue_variant
              _nef_sequence.cis_peptide

             1    A   20   GLU   start    .   .
             2    A   21   TYR   middle   .   .
             3    A   22   ALA   middle   .   .
             4    A   23   GLN   middle   .   .
             5    A   24   PRO   middle   .   .
             6    A   25   ARG   middle   .   .
             7    A   26   LEU   middle   .   .
             8    A   27   ARG   middle   .   .
             9    A   28   LEU   middle   .   .
             10   A   29   GLY   middle   .   .
             11   A   30   PHE   middle   .   .
             12   A   31   GLU   middle   .   .
             13   A   32   ASP   end      .   .

           stop_

        save_
    """

    assert_lines_match(EXPECTED, result.stdout)


# noinspection PyUnusedLocal
def test_renumber_multi_chain_two_chains():
    path = path_in_test_data(__file__, "multi_chain.nef")

    result = run_and_report(app, [*OFFSET_CHAINS_A_10_B_5], input=open(path))

    EXPECTED = """\
        data_test

        save_nef_molecular_system
           _nef_molecular_system.sf_category   nef_molecular_system
           _nef_molecular_system.sf_framecode  nef_molecular_system

            loop_
                _nef_sequence.index
                _nef_sequence.chain_code
                _nef_sequence.sequence_code
                _nef_sequence.residue_name
                _nef_sequence.linking
                _nef_sequence.residue_variant
                _nef_sequence.cis_peptide
                _nef_sequence.ccpn_comment
                _nef_sequence.ccpn_chain_role
                _nef_sequence.ccpn_compound_name
                _nef_sequence.ccpn_chain_comment
                1   A   13   HIS   .   .   .   .   .   Sec5   .
                2   A   14   MET   .   .   .   .   .   Sec5   .
                3   B   10   ARG   .   .   .   .   .   Sec5   .
                4   B   11   GLN   .   .   .   .   .   Sec5   .
                5   C   7    PRO   .   .   .   .   .   Sec5   .
            stop_

        save_
    """

    assert_lines_match(EXPECTED, result.stdout)


# noinspection PyUnusedLocal
def test_renumber_no_offset():
    path = path_in_test_data(__file__, "multi_chain.nef")

    result = run_and_report(
        app, [*OFFSET_CHAINS_MISSING_OFFSET], input=open(path), expected_exit_code=1
    )

    assert "ERROR" in result.stdout
    assert "there must be an offset/start for each chain" in result.stdout


# noinspection PyUnusedLocal
def test_renumber_bad_offset():
    path = path_in_test_data(__file__, "multi_chain.nef")

    result = run_and_report(
        app, [*OFFSET_CHAINS_BAD_OFFSET], input=open(path), expected_exit_code=1
    )

    assert "ERROR" in result.stdout
    assert "B5" in result.stdout
    assert "2nd" in result.stdout
    assert "can't be converted to an int" in result.stdout


def test_renumber_no_chain_or_offset():
    path = path_in_test_data(__file__, "multi_chain.nef")

    result = run_and_report(app, [], input=open(path), expected_exit_code=1)

    assert "ERROR" in result.stdout
    assert "you didn't provide any chains and offsets/starts" in result.stdout


# noinspection PyUnusedLocal
def test_set_starts_two_chains():
    path = path_in_test_data(__file__, "multi_chain.nef")

    result = run_and_report(app, [*SET_CHAIN_STARTS_A_0_B_10], input=open(path))

    EXPECTED = """\
        data_test

        save_nef_molecular_system
           _nef_molecular_system.sf_category   nef_molecular_system
           _nef_molecular_system.sf_framecode  nef_molecular_system

           loop_
              _nef_sequence.index
              _nef_sequence.chain_code
              _nef_sequence.sequence_code
              _nef_sequence.residue_name
              _nef_sequence.linking
              _nef_sequence.residue_variant
              _nef_sequence.cis_peptide
              _nef_sequence.ccpn_comment
              _nef_sequence.ccpn_chain_role
              _nef_sequence.ccpn_compound_name
              _nef_sequence.ccpn_chain_comment

             1   A   0    HIS   .   .   .   .   .   Sec5   .
             2   A   1    MET   .   .   .   .   .   Sec5   .
             3   B   10   ARG   .   .   .   .   .   Sec5   .
             4   B   11   GLN   .   .   .   .   .   Sec5   .
             5   C   7    PRO   .   .   .   .   .   Sec5   .

           stop_

        save_

    """

    assert_lines_match(EXPECTED, result.stdout)


def test_renumber_shifts_chain_a_only():
    path = path_in_test_data(__file__, "nef_3_peaks.nef")

    result = run_and_report(app, [*SET_CHAIN_STAR_A_SHIFTS], input=open(path))

    NOQUA_E501 = "# noqa E501"
    EXPECTED = """# noqa E501
        data_nmrview

        save_nef_molecular_system
           _nef_molecular_system.sf_category   nef_molecular_system
           _nef_molecular_system.sf_framecode  nef_molecular_system

           loop_
              _nef_sequence.index
              _nef_sequence.chain_code
              _nef_sequence.sequence_code
              _nef_sequence.residue_name
              _nef_sequence.linking
              _nef_sequence.residue_variant
              _nef_sequence.cis_peptide

             1   A   1   ALA   start    .   .
             2   A   2   ALA   middle   .   .
             3   A   3   ALA   end      .   .

           stop_

        save_

        save_nef_nmr_spectrum_simnoe
           _nef_nmr_spectrum.sf_category          nef_nmr_spectrum
           _nef_nmr_spectrum.sf_framecode         nef_nmr_spectrum_simnoe
           _nef_nmr_spectrum.num_dimensions       3
           _nef_nmr_spectrum.chemical_shift_list  .

           loop_
              _nef_spectrum_dimension.dimension_id
              _nef_spectrum_dimension.axis_unit
              _nef_spectrum_dimension.axis_code
              _nef_spectrum_dimension.spectrometer_frequency
              _nef_spectrum_dimension.spectral_width
              _nef_spectrum_dimension.value_first_point
              _nef_spectrum_dimension.folding
              _nef_spectrum_dimension.absolute_peak_positions
              _nef_spectrum_dimension.is_acquisition

             1   ppm   1H    800.133   11.0113   .   circular   true   .
             2   ppm   1H    800.133   11.0600   .   circular   true   .
             3   ppm   15N   81.076    81.8927   .   circular   true   .

           stop_

           loop_
              _nef_spectrum_dimension_transfer.dimension_1
              _nef_spectrum_dimension_transfer.dimension_2
              _nef_spectrum_dimension_transfer.transfer_type


           stop_

           loop_
              _nef_peak.index
              _nef_peak.peak_id
              _nef_peak.volume
              _nef_peak.volume_uncertainty
              _nef_peak.height
              _nef_peak.height_uncertainty
              _nef_peak.position_1
              _nef_peak.position_uncertainty_1
              _nef_peak.position_2
              _nef_peak.position_uncertainty_2
              _nef_peak.position_3
              _nef_peak.position_uncertainty_3
              _nef_peak.chain_code_1
              _nef_peak.sequence_code_1
              _nef_peak.residue_name_1
              _nef_peak.atom_name_1
              _nef_peak.chain_code_2
              _nef_peak.sequence_code_2
              _nef_peak.residue_name_2
              _nef_peak.atom_name_2
              _nef_peak.chain_code_3
              _nef_peak.sequence_code_3
              _nef_peak.residue_name_3
              _nef_peak.atom_name_3

             1   0   0.38    .   0.38    .   10.405   .   8.796   .   132.49   .   A   10   ala   HE1   A   12   ala   HN   A   10   ala   N
             2   1   1.298   .   1.298   .   10.408   .   7.139   .   132.49   .   .   .    .     .     .   .    .     .    .   .    .     .
             3   3   0.319   .   0.319   .   10.408   .   5.542   .   132.49   .   A   11   ala   HE1   A   12   ala   HA   .   .    .     .

           stop_

        save_


    """.replace(
        NOQUA_E501, ""
    )

    assert_lines_match(EXPECTED, result.stdout)



import typer
from typer.testing import CliRunner

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.tools.chains.rename import rename

runner = CliRunner()

app = typer.Typer()
app.command()(rename)

RENAME_CHAINS_B_D = ["B", "D"]
RENAME_CHAINS_A_E = ["A", "E"]


# noinspection PyUnusedLocal
def test_rename_basic(clear_cache):

    path = path_in_test_data(__file__, "multi_chain.nef")

    result = run_and_report(app, [*RENAME_CHAINS_B_D], input=open(path))

    if result.exit_code != 0:
        print("INFO: stdout from failed read:\n", result.stdout)

    assert result.exit_code == 0

    EXPECTED = """\
    data_test

    save_nef_molecular_system
       _nef_molecular_system.sf_category   nef_molecular_system
       _nef_molecular_system.sf_framecode  nef_molecular_system

       loop_
          _nef_sequence.index
          _nef_sequence.chain_code
          _nef_sequence.sequence_code
          _nef_sequence.residue_name
          _nef_sequence.linking
          _nef_sequence.residue_variant
          _nef_sequence.cis_peptide
          _nef_sequence.ccpn_comment
          _nef_sequence.ccpn_chain_role
          _nef_sequence.ccpn_compound_name
          _nef_sequence.ccpn_chain_comment

         1   A   3   HIS   .   .   .   .   .   Sec5   .
         2   A   4   MET   .   .   .   .   .   Sec5   .
         3   D   5   ARG   .   .   .   .   .   Sec5   .
         4   D   6   GLN   .   .   .   .   .   Sec5   .
         5   C   7   PRO   .   .   .   .   .   Sec5   .

       stop_

    save_

    """

    assert_lines_match(EXPECTED, result.stdout)


# noinspection PyUnusedLocal
def test_rename_multi_frame(clear_cache):

    path = path_in_test_data(__file__, "multi_chain_shifts.nef")

    result = run_and_report(app, [*RENAME_CHAINS_A_E], input=open(path))

    if result.exit_code != 0:
        print("INFO: stdout from failed read:\n", result.stdout)

    assert result.exit_code == 0

    EXPECTED = """\
        data_test

        save_nef_molecular_system
           _nef_molecular_system.sf_category   nef_molecular_system
           _nef_molecular_system.sf_framecode  nef_molecular_system

           loop_
              _nef_sequence.index
              _nef_sequence.chain_code
              _nef_sequence.sequence_code
              _nef_sequence.residue_name
              _nef_sequence.linking
              _nef_sequence.residue_variant
              _nef_sequence.cis_peptide
              _nef_sequence.ccpn_comment
              _nef_sequence.ccpn_chain_role
              _nef_sequence.ccpn_compound_name
              _nef_sequence.ccpn_chain_comment

             1   E   3   HIS   .   .   .   .   .   Sec5   .
             2   E   4   MET   .   .   .   .   .   Sec5   .
             3   B   5   ARG   .   .   .   .   .   Sec5   .
             4   B   6   GLN   .   .   .   .   .   Sec5   .
             5   C   7   PRO   .   .   .   .   .   Sec5   .

           stop_

        save_

        save_ccpn_residue_assignments_test
            _ccpn_residue_assignments_test.sf_category                      ccpn_residue_assignments
            _ccpn_residue_assignments_test.sf_frame_code                    ccpn_residue_assignments_test
            _ccpn_residue_assignments_test.ccpn_assignment_program          mars
            _ccpn_residue_assignments_test.ccpn_assignment_program_version  .
            _ccpn_residue_assignments_test.ccpn_assignment_source           .

           loop_
              _ccpn_residue_assignments_test.serial
              _ccpn_residue_assignments_test.chain_code
              _ccpn_residue_assignments_test.residue_number
              _ccpn_residue_assignments_test.residue_type
              _ccpn_residue_assignments_test.assignment_serial
              _ccpn_residue_assignments_test.assignment
              _ccpn_residue_assignments_test.merit
              _ccpn_residue_assignments_test.fixed

             0   E   1   HIS   .   .    .      .
             1   E   2   MET   0   30   0.2    False
             1   E   2   MET   1   48   0.18   False
             1   B   2   MET   2   50   0.13   False
             1   B   2   MET   3   51   0.15   False
             1   B   2   MET   4   56   0.15   False

           stop_

        save_

    """

    assert_lines_match(EXPECTED, result.stdout)


# noinspection PyUnusedLocal
def test_rename_select_frame_by_name(clear_cache):

    path = path_in_test_data(__file__, "multi_chain_shifts.nef")

    result = run_and_report(
        app, ["--frame", "test", *RENAME_CHAINS_A_E], input=open(path)
    )

    if result.exit_code != 0:
        print("INFO: stdout from failed read:\n", result.stdout)

    assert result.exit_code == 0

    EXPECTED = """\
        data_test

        save_nef_molecular_system
           _nef_molecular_system.sf_category   nef_molecular_system
           _nef_molecular_system.sf_framecode  nef_molecular_system

           loop_
              _nef_sequence.index
              _nef_sequence.chain_code
              _nef_sequence.sequence_code
              _nef_sequence.residue_name
              _nef_sequence.linking
              _nef_sequence.residue_variant
              _nef_sequence.cis_peptide
              _nef_sequence.ccpn_comment
              _nef_sequence.ccpn_chain_role
              _nef_sequence.ccpn_compound_name
              _nef_sequence.ccpn_chain_comment

             1   A   3   HIS   .   .   .   .   .   Sec5   .
             2   A   4   MET   .   .   .   .   .   Sec5   .
             3   B   5   ARG   .   .   .   .   .   Sec5   .
             4   B   6   GLN   .   .   .   .   .   Sec5   .
             5   C   7   PRO   .   .   .   .   .   Sec5   .
           stop_
        save_

        save_ccpn_residue_assignments_test
            _ccpn_residue_assignments_test.sf_category                      ccpn_residue_assignments
            _ccpn_residue_assignments_test.sf_frame_code                    ccpn_residue_assignments_test
            _ccpn_residue_assignments_test.ccpn_assignment_program          mars
            _ccpn_residue_assignments_test.ccpn_assignment_program_version  .
            _ccpn_residue_assignments_test.ccpn_assignment_source           .

            loop_
                _ccpn_residue_assignments_test.serial
                _ccpn_residue_assignments_test.chain_code
                _ccpn_residue_assignments_test.residue_number
                _ccpn_residue_assignments_test.residue_type
                _ccpn_residue_assignments_test.assignment_serial
                _ccpn_residue_assignments_test.assignment
                _ccpn_residue_assignments_test.merit
                _ccpn_residue_assignments_test.fixed

                 0   E   1   HIS   .   .    .      .
                 1   E   2   MET   0   30   0.2    False
                 1   E   2   MET   1   48   0.18   False
                 1   B   2   MET   2   50   0.13   False
                 1   B   2   MET   3   51   0.15   False
                 1   B   2   MET   4   56   0.15   False
           stop_
        save_

    """

    assert_lines_match(EXPECTED, result.stdout)


# noinspection PyUnusedLocal
def test_rename_select_frame_by_category(clear_cache):

    path = path_in_test_data(__file__, "multi_chain_shifts.nef")
    input = open(path).read()

    command = [*RENAME_CHAINS_A_E, "--frame", "mol", "--selector-type", "category"]

    result = run_and_report(app, command, input=input)

    if result.exit_code != 0:
        print("INFO: stdout from failed read:\n", result.stdout)

    assert result.exit_code == 0

    EXPECTED = """\
    data_test

        save_nef_molecular_system
           _nef_molecular_system.sf_category   nef_molecular_system
           _nef_molecular_system.sf_framecode  nef_molecular_system

           loop_
              _nef_sequence.index
              _nef_sequence.chain_code
              _nef_sequence.sequence_code
              _nef_sequence.residue_name
              _nef_sequence.linking
              _nef_sequence.residue_variant
              _nef_sequence.cis_peptide
              _nef_sequence.ccpn_comment
              _nef_sequence.ccpn_chain_role
              _nef_sequence.ccpn_compound_name
              _nef_sequence.ccpn_chain_comment

             1   E   3   HIS   .   .   .   .   .   Sec5   .
             2   E   4   MET   .   .   .   .   .   Sec5   .
             3   B   5   ARG   .   .   .   .   .   Sec5   .
             4   B   6   GLN   .   .   .   .   .   Sec5   .
             5   C   7   PRO   .   .   .   .   .   Sec5   .
           stop_
        save_

        save_ccpn_residue_assignments_test
            _ccpn_residue_assignments_test.sf_category                      ccpn_residue_assignments
            _ccpn_residue_assignments_test.sf_frame_code                    ccpn_residue_assignments_test
            _ccpn_residue_assignments_test.ccpn_assignment_program          mars
            _ccpn_residue_assignments_test.ccpn_assignment_program_version  .
            _ccpn_residue_assignments_test.ccpn_assignment_source           .

            loop_
                _ccpn_residue_assignments_test.serial
                _ccpn_residue_assignments_test.chain_code
                _ccpn_residue_assignments_test.residue_number
                _ccpn_residue_assignments_test.residue_type
                _ccpn_residue_assignments_test.assignment_serial
                _ccpn_residue_assignments_test.assignment
                _ccpn_residue_assignments_test.merit
                _ccpn_residue_assignments_test.fixed

                 0   A  1   HIS   .   .    .      .
                 1   A   2   MET   0   30   0.2    False
                 1   A   2   MET   1   48   0.18   False
                 1   B   2   MET   2   50   0.13   False
                 1   B   2   MET   3   51   0.15   False
                 1   B   2   MET   4   56   0.15   False
           stop_
        save_

    """

    assert_lines_match(EXPECTED, result.stdout)



import sys
from dataclasses import dataclass
from enum import auto
from pathlib import Path
from typing import List, Tuple

import typer
from pynmrstar import Entry
from strenum import LowercaseStrEnum

from nef_pipelines.lib.isotope_lib import Isotope
from nef_pipelines.lib.nef_lib import (
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames_by_name,
)
from nef_pipelines.lib.peak_lib import peaks_to_frame
from nef_pipelines.lib.shift_lib import nef_frames_to_shifts
from nef_pipelines.lib.structures import NewPeak, ShiftData
from nef_pipelines.lib.util import FStringTemplate, parse_comma_separated_options
from nef_pipelines.tools.shifts import shifts_app


class SpectraTypes(LowercaseStrEnum):
    N_HSQC = auto()
    C_HSQC = auto()

    HNCA = auto()
    HNcoCA = auto()

    HNCACB = auto()
    HNcoCACB = auto()
    CBCAcoNH = auto()

    HNCO = auto()
    HNcaCO = auto()

    TRIPLE = auto()


TRIPLE_SPECTRA = {
    SpectraTypes.N_HSQC,
    SpectraTypes.HNCA,
    SpectraTypes.HNcoCA,
    SpectraTypes.HNCACB,
    SpectraTypes.HNcoCACB,
    SpectraTypes.HNCO,
    SpectraTypes.HNcaCO,
}
#

ATOMS_TO_DIMENSIONS = {"H", "1H"}
ATOMS = "atoms"
DIMENSIONS = "dimensions"
RESIDUE_OFFSETS = "residue_offsets"


@dataclass
class ShiftInfo:
    atoms: List[str]
    dimensions: List[Isotope]
    atom_sets: List[List[Tuple[int, str]]]
    atom_set_signs: List[int]


@dataclass
class AtomInfo:
    n_field: int
    atom_type: str


# TODO much of this is ugly!
SPECTRA_INFO = {
    SpectraTypes.N_HSQC: ShiftInfo(
        atoms={"H", "N"},
        dimensions=(Isotope.H1, Isotope.N15),
        atom_sets=[
            [(1, "H"), (1, "N")],
        ],
        atom_set_signs=[1],
    ),
    # SpectraTypes.C_HSQC: ShiftInfo(
    #     atoms = {'H', 'C'},
    #     dimensions= (Isotope.H1, Isotope.N15),
    #     atom_sets = [[(1, 'H'), (1, 'C')],],
    #     atom_set_signs=[1]
    # ),
    SpectraTypes.HNCA: ShiftInfo(
        atoms={"H", "N", ("CA",)},
        dimensions=(Isotope.H1, Isotope.N15, Isotope.C13),
        atom_sets=[
            [(1, "H"), (1, "N"), (1, "CA")],
            [(1, "H"), (1, "N"), (2, "CA")],
        ],
        atom_set_signs=[1, 1],
    ),
    SpectraTypes.HNcoCA: ShiftInfo(
        atoms={"H", "N", ("CA",)},
        dimensions=(Isotope.H1, Isotope.N15, Isotope.C13),
        atom_sets=[
            [(1, "H"), (1, "N"), (2, "CA")],
        ],
        atom_set_signs=[1],
    ),
    SpectraTypes.HNCACB: ShiftInfo(
        atoms={"H", "N", ("CA", "CB")},
        dimensions=(Isotope.H1, Isotope.N15, Isotope.C13),
        atom_sets=[
            [(1, "H"), (1, "N"), (1, "CA")],
            [(1, "H"), (1, "N"), (2, "CA")],
            [(1, "H"), (1, "N"), (1, "CB")],
            [(1, "H"), (1, "N"), (2, "CB")],
        ],
        atom_set_signs=[1, 1, -1, -1],
    ),
    SpectraTypes.HNcoCACB: ShiftInfo(
        atoms={"H", "N", ("CA", "CB")},
        dimensions=(Isotope.H1, Isotope.N15, Isotope.C13),
        atom_sets=[
            [(1, "H"), (1, "N"), (2, "CA")],
            [(1, "H"), (1, "N"), (2, "CB")],
        ],
        atom_set_signs=[1, -1],
    ),
    SpectraTypes.CBCAcoNH: ShiftInfo(
        atoms={"H", "N", ("CA", "CB")},
        dimensions=(Isotope.H1, Isotope.N15, Isotope.C13),
        atom_sets=[
            [(1, "H"), (1, "N"), (2, "CA")],
            [(1, "H"), (1, "N"), (2, "CB")],
        ],
        atom_set_signs=[1, 1],
    ),
    SpectraTypes.HNCO: ShiftInfo(
        atoms={"H", "N", ("C",)},
        dimensions=(Isotope.H1, Isotope.N15, Isotope.C13),
        atom_sets=([(1, "H"), (1, "N"), (2, "C")],),
        atom_set_signs=[1],
    ),
    SpectraTypes.HNcaCO: ShiftInfo(
        atoms={"H", "N", ("C",)},
        dimensions=(Isotope.H1, Isotope.N15, Isotope.C13),
        atom_sets=[
            [(1, "H"), (1, "N"), (1, "C")],
            [(1, "H"), (1, "N"), (2, "C")],
        ],
        atom_set_signs=[1, 1],
    ),
}


SHIFT_FRAMES_HELP = """\
    the names of the shift frames to use, this can be a comma separated list of name or the option can be called
    called multiple times. Wild cards are allowed. If no match is found wild cards are checked as well unless
    --exact is set
"""

SPECTRA_HELP = f"""
    the names of the spectra to create, this can be a comma separated list of options or
    called multiple times. Possible values are: {', '.join(SpectraTypes)}
"""


@shifts_app.command()
def make_peaks(
    shift_frame_selectors: List[str] = typer.Option(
        ["default"], help=SHIFT_FRAMES_HELP
    ),
    exact: bool = typer.Option(
        False, help="if set frames are selected by exact matches"
    ),
    spectra: List[str] = typer.Option([SpectraTypes.N_HSQC], help=SPECTRA_HELP),
    name_template: str = typer.Option("synthetic_{spectrum}"),
    spectrometer_frequency: float = typer.Option("600.12345678"),
    input: Path = typer.Option(
        Path("-"),
        "--input",
        "-i",
        metavar="|INPUT|",
        help="input to read NEF data from [stdin = -]",
    ),
):
    """-  make a set of peaks for triple resonance and hsqc spectra from a list of shifts [alpha!]"""

    print(
        "*** WARNING *** this command [shifts make_peaks] is only lightly tested use at your own risk!",
        file=sys.stderr,
    )

    shift_frames = parse_comma_separated_options(shift_frame_selectors)

    spectra = set(parse_comma_separated_options(spectra))

    spectra = _update_spectra_with_groups(spectra)

    # TODO add a check that we have some shift frames

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    entry = pipe(
        entry, shift_frames, exact, spectra, name_template, spectrometer_frequency
    )

    print(entry)


def _update_spectra_with_groups(spectra):
    if SpectraTypes.TRIPLE in spectra:
        spectra.update(TRIPLE_SPECTRA)
        spectra.remove(SpectraTypes.TRIPLE)

    return spectra


def pipe(
    entry: Entry,
    shift_frame_selectors: List[str],
    exact_frame_selectors: bool,
    spectra: List[SpectraTypes],
    name_template_string: str,
    spectrometer_frequency: float = 600.123456789,
) -> Entry:
    all_shift_list_frames = entry.get_saveframes_by_category("nef_chemical_shift_list")

    selected_shift_list_frames = select_frames_by_name(
        all_shift_list_frames, shift_frame_selectors, exact=exact_frame_selectors
    )

    shifts = nef_frames_to_shifts(selected_shift_list_frames)

    name_template = FStringTemplate(name_template_string)

    for spectrum in spectra:

        frame_name = str(name_template)

        spectrum_info = SPECTRA_INFO[spectrum]
        peaks = make_spectrum(shifts, spectrum_info)

        dimensions = [
            {"axis_code": dimension} for dimension in spectrum_info.dimensions
        ]

        # TODO add info about frames for errors
        frame = peaks_to_frame(peaks, dimensions, spectrometer_frequency, frame_name)

        entry.add_saveframe(frame)

    return entry


def make_spectrum(
    shifts: List[ShiftData], info: ShiftInfo, height: int = 1_000_000
) -> List[NewPeak]:

    # note this is not very sophisticated and doesn't cover side-chains... we ought to port the ccpn code
    relevant_shifts = []

    for atom_name in info.atoms:
        for shift in shifts:
            shift_atom_name = shift.atom.atom_name.upper()
            shift_atom_name = shift_atom_name.strip("1")
            shift_atom_name = shift_atom_name.strip("-")

            if shift_atom_name in atom_name:
                relevant_shifts.append(shift)

    shifts_by_residue = {}

    for shift in relevant_shifts:
        residue = str(shift.atom.residue.sequence_code).split("-")[0]
        shifts_by_residue.setdefault(residue, []).append(shift)

    peaks = []

    for residue_shifts in shifts_by_residue.values():

        for atom_set, sign in zip(info.atom_sets, info.atom_set_signs):

            atom_set_shifts = []
            for number_sequence_code_fields_required, atom_name_required in atom_set:
                for shift in residue_shifts:

                    # these are errors for the spectra we are interested in currently
                    # maybe add a warning?
                    if "+" in str(shift.atom.residue.sequence_code):
                        continue

                    number_sequence_code_fields = len(
                        str(shift.atom.residue.sequence_code).split("-")
                    )
                    atom_name = shift.atom.atom_name

                    number_fields_ok = (
                        number_sequence_code_fields
                        == number_sequence_code_fields_required
                    )
                    atom_name_ok = atom_name_required in atom_name

                    if number_fields_ok and atom_name_ok:
                        atom_set_shifts.append(shift)

            if len(atom_set_shifts) == len(info.atoms):
                new_peak = NewPeak(
                    atom_set_shifts, height=height * sign, volume=height * sign
                )
                peaks.append(new_peak)

    return peaks


# def make_triple_resonance(shifts: List[ShiftData], info: Dict[str, Any], height: int =1_000_000) -> List[NewPeak]:
#
#     # note this is not very sophisticated and doesn't cover sidechains... we ought to port the ccpn code
#     relevant_shifts = []
#
#
#     for atom_name in info.atom_names:
#         for shift in shifts:
#             if shift.atom.atom_name.upper() in atom_name:
#                 relevant_shifts.append(shift)
#
#
#     shifts_by_residue = {}
#
#     for shift in relevant_shifts:
#         residue = shift.atom.residue.sequence_code.split('-')[0]
#         shifts_by_residue.setdefault(residue, []).append(shift)
#
#     print(shifts_by_residue)
#
#     # for shift in hsqc_shifts:
#     #     shifts_by_residue.setdefault(shift.atom.residue, []).append(shift)
#     #
#     # peaks = []
#     # for shifts_by_residue in shifts_by_residue.values():
#     #     if len(shifts_by_residue) != 2:
#     #         continue
#     #
#     #     dimension_shifts = []
#     #     for atom_name in atom_names:
#     #         for shift in shifts_by_residue:
#     #             if shift.atom.atom_name == atom_name:
#     #                 dimension_shifts.append(shift)
#     #
#     #     if len(dimension_shifts) != 2:
#     #         continue
#     #
#     #     peaks.append(NewPeak(dimension_shifts, height=height))
#
#     return peaks



import typer
from typer.testing import CliRunner

from nef_pipelines.lib.nef_lib import NEF_MOLECULAR_SYSTEM
from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.fasta.importers.sequence import sequence

runner = CliRunner()
app = typer.Typer()
app.command()(sequence)


EXPECTED_3AA = """\
save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1   A   1   ALA   start    .   .
     2   A   2   ALA   middle   .   .
     3   A   3   ALA   end      .   .

   stop_

save_"""


# noinspection PyUnusedLocal
def test_3aa(clear_cache):

    path = path_in_test_data(__file__, "3aa.fasta")
    result = run_and_report(app, [path])

    mol_sys_result = isolate_frame(result.stdout, "%s" % NEF_MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3AA, mol_sys_result)


EXPECTED_3A_AB = """\
save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1   A    1   ALA   start    .   .
     2   A    2   ALA   middle   .   .
     3   A    3   ALA   end      .   .
     4   B    1   ALA   start    .   .
     5   B    2   ALA   middle   .   .
     6   B    3   ALA   end      .   .

   stop_

save_"""


# noinspection PyUnusedLocal
def test_3aa_x2(clear_cache):

    path = path_in_test_data(__file__, "3aa_x2.fasta")
    result = run_and_report(app, [path])

    assert result.exit_code == 0

    mol_sys_result = isolate_frame(result.stdout, "%s" % NEF_MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3A_AB, mol_sys_result)


EXPECTED_3A_AB_B_start_11 = """\
save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1   A    1   ALA   start    .   .
     2   A    2   ALA   middle   .   .
     3   A    3   ALA   end      .   .
     4   B   11   ALA   start    .   .
     5   B   12   ALA   middle   .   .
     6   B   13   ALA   end      .   .

   stop_

save_"""


# noinspection PyUnusedLocal
def test_3aa_x2_off_10_b(clear_cache):

    path = path_in_test_data(__file__, "3aa_x2.fasta")
    result = run_and_report(app, ["--starts", "1,11", path])

    mol_sys_result = isolate_frame(result.stdout, "%s" % NEF_MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3A_AB_B_start_11, mol_sys_result)



import sys
import traceback
from fnmatch import fnmatch
from io import StringIO
from itertools import zip_longest
from pathlib import Path
from typing import IO, AnyStr, List, Optional

from click.testing import Result
from pynmrstar import Entry
from typer import Typer
from typer.testing import CliRunner

NOQA_E501 = "# noqa: E501"


def run_and_read_pytest(args):

    from pytest import main

    original_output = sys.stdout
    original_error = sys.stdout
    sys.stdout = StringIO()
    sys.stderr = StringIO()

    retcode = main(args)

    output = sys.stdout.getvalue()
    error_output = sys.stderr.getvalue()

    sys.stdout.close()
    sys.stderr.close()
    sys.stdout = original_output
    sys.stderr = original_error

    return retcode, output, error_output


def _split_test_spec(spec):
    spec_parts = spec.split("::")
    spec_parts[0] = Path(spec_parts[0])

    return spec_parts


def select_matching_tests(tests, selectors):
    results = []

    for selector in selectors:

        selector_parts = _split_test_spec(selector)
        num_selector_parts = len(selector_parts)

        if num_selector_parts == 1:
            selector = f"*::{selector_parts[0]}"
            selector_parts = _split_test_spec(selector)
            num_selector_parts = len(selector_parts)

        if num_selector_parts == 2 and selector_parts[0] == Path(""):
            selector = f"*::{selector_parts[1]}"
            selector_parts = _split_test_spec(selector)
            num_selector_parts = len(selector_parts)

        if num_selector_parts == 2 and selector_parts[1] == "":
            selector = f"{selector_parts[0]}::*"
            selector_parts = _split_test_spec(selector)
            num_selector_parts = len(selector_parts)

        selector_path_parts = selector_parts[0].parts
        num_selector_path_parts = len(selector_path_parts)

        for test in tests:

            # here we ensure we are looking for a .py file...
            if not selector_path_parts[-1].endswith("py"):
                selector_path_parts = list(selector_path_parts)
                selector_path_parts[-1] = f"{selector_path_parts[-1]}.py"
            selector_path_parts = tuple(selector_path_parts)

            test_parts = _split_test_spec(test)
            num_test_parts = len(test_parts)
            test_path_parts = test_parts[0].parts
            num_test_path_parts = len(test_path_parts)

            paths_equal = False
            if num_selector_path_parts <= num_test_path_parts:
                selector_path = selector_path_parts[-num_selector_path_parts:]
                test_path = test_path_parts[-num_selector_path_parts:]
                paths_equal = selector_path == test_path

            path = str(test_parts[0])
            path_test = str(selector_parts[0])

            if not paths_equal:
                paths_equal = fnmatch(path, path_test)

            test_names_equal = False
            if (num_selector_parts == 2) and (num_test_parts == 2):
                test_names_equal = fnmatch(test_parts[-1], selector_parts[-1])

            if paths_equal and test_names_equal:
                results.append(test)
    return results


def assert_lines_match(
    expected: str, reported: str, squash_spaces: bool = True, ignore_empty=True
):
    """
    compare two multi line strings line by line with stripping and raise an assertion if they don't match
    note: empty lines are ignoresd by default, and multiple spaces are squashed
    Args:
        expected (str): the expected string
        reported (str): the input string
        squash_spaces (bool): remove duplicate spaces before comparison

    Returns:
        None
    """
    lines_expected = expected.split("\n")
    lines_reported = reported.split("\n")

    if ignore_empty:
        lines_expected = [line for line in lines_expected if len(line.strip()) != 0]
        lines_reported = [line for line in lines_reported if len(line.strip()) != 0]

    zip_lines = zip_longest(lines_expected, lines_reported, fillvalue="")
    for i, (expected_line, reported_line) in enumerate(zip_lines):

        expected_line_stripped = expected_line.strip()
        reported_line_stripped = reported_line.strip()

        if squash_spaces:
            expected_line_stripped = " ".join(expected_line_stripped.split())
            reported_line_stripped = " ".join(reported_line_stripped.split())

        if reported_line_stripped != expected_line_stripped:

            for line in lines_expected:
                print(f"exp|{line}")
            print()

            for line in lines_reported:
                print(f"rep|{line}")
            print()

            print("line that caused the error:")
            print()

            print(f"exp|{i}|  |{expected_line_stripped}|")
            print(f"rep|{i}|  |{reported_line_stripped}|")

        assert reported_line_stripped == expected_line_stripped


def isolate_frame(target: str, name: str) -> Optional[str]:
    """
    Extract one frame from a NEF file by name as a string
    Args:
        target (Entry): target NEF entry
        name (str): name of the save frame

    Returns:
        Optional[str]: the entry or a None if not found
    """

    entry = Entry.from_string(target)
    entry = str(entry.get_saveframe_by_name(name))

    return entry


def path_in_parent_directory(root: str, target: str) -> str:
    """
    given a root and a relative path find the relative file path

    Args:
        root (str): root of the path
        target (str): the relative path from the root

    Returns:
        str: the target paths as a string
    """
    parent_path = Path(root).parent
    return str(Path(parent_path, target).absolute())


def root_path(initial_path: str):
    """given a path work up the directory structure till you find the
    directory containing the nef executable

    initial_path (str): the path to start searching from
    """

    target = Path(initial_path)

    belt_and_braces = 100  # noqa: F841 this appears to be a bug
    while (
        not (Path(target.root) == target)
        and not (target / "nef_pipelines" / "main.py").is_file()
    ):
        target = target.parent
        belt_and_braces -= 1
        if belt_and_braces < 0:
            msg = f"""\
                Error, while search for the rot of the path {initial_path} i walked up 100
                directories this looks like a bug!
            """
            raise Exception(msg)

    target /= "nef_pipelines"

    return target


# TODO: remove local we now use a hierarchical search
def path_in_test_data(root: str, file_name: str) -> str:
    """
    given a root and a file name find the relative to the file
    in the parents test_data directory

    Args:
        root (str): root of the path
        file_name (str): the name of the file

    Returns:
        str: the target paths as a string
    """

    test_data_root = root_path(root) / "tests" / "test_data"
    test_data_local = path_in_parent_directory(root, "test_data")

    if (Path(test_data_local) / file_name).is_file():
        test_data = test_data_local
    else:
        test_data = test_data_root

    # TODO: add an error on shadowing global files

    return str(Path(test_data, file_name).absolute())


def run_and_report(
    typer_app: Typer,
    args: List[str],
    input: IO[AnyStr] = None,
    expected_exit_code: int = 0,
) -> Result:
    """
    run a typer app in the typer test harness and report exceptions and stdout to screen if there is an error
    :param typer_app: the typer app hosting the application
    :param args: command line arguments for the app
    :param input: an input stream if required
    :param expected_exit_code: what exit code to expect if the app is expected to end with an error
    :return: results object
    """

    runner = CliRunner()
    result = runner.invoke(typer_app, args, input=input)

    if result.exit_code != expected_exit_code:
        print("\n", "-" * 40, "-stdout-", "-" * 40)
        print(result.stdout)
        if result.exception:
            print("-" * 40, "exception", "-" * 40)
            formatted = list(traceback.TracebackException(*result.exc_info).format())

            # TODO: this is a hack, I would love to have a better solution!
            if "SystemExit" in formatted[-1]:
                for i, line in enumerate(reversed(formatted)):
                    if (
                        "During handling of the above exception, another exception occurred"
                        in line
                    ):
                        break
                formatted = formatted[: -i - 2]
            print("".join(formatted))

        print("-" * 40, "-" * 9, "-" * 40)

    assert result.exit_code == expected_exit_code

    return result



import typer
from typer.testing import CliRunner

from nef_pipelines.lib.nef_lib import NEF_MOLECULAR_SYSTEM
from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.xplor.importers.sequence import sequence

runner = CliRunner()
app = typer.Typer()
app.command()(sequence)

EXPECTED_3A_AB = """\
save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1   AAAA  1   ALA   start    .   .
     2   AAAA  2   ALA   middle   .   .
     3   AAAA  3   ALA   end      .   .
     4   BBBB  11  ALA   start    .   .
     5   BBBB  12  ALA   middle   .   .
     6   BBBB  13  ALA   end      .   .

   stop_

save_"""


# noinspection PyUnusedLocal
def test_3ab(clear_cache):

    path = path_in_test_data(__file__, "3a_ab.psf")
    result = run_and_report(app, [path])

    result = isolate_frame(result.stdout, "%s" % NEF_MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3A_AB, result)


def test_no_sequence_files(clear_cache):

    result = run_and_report(app, [], expected_exit_code=1)

    assert "no psf files provided to read sequences from" in result.stdout



from dataclasses import replace
from pathlib import Path
from typing import List, Tuple

import tabulate
import typer
from pynmrstar import Saveframe

from nef_pipelines.lib.nef_lib import (
    loop_row_dict_iter,
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames_by_name,
)
from nef_pipelines.lib.structures import AtomLabel, RdcRestraint, SequenceResidue
from nef_pipelines.transcoders.xplor import export_app

app = typer.Typer()

TENSOR_ATOMS_HELP = """\
    name of the atoms in the tensor in the order o,z,x,y can be provided as a series of calls or
    or comma separated values with no gaps
"""

TENSOR_CHAIN_CODE_HELP = (
    """chain_code/segid of the tensor frame [taken from nef input if present]"""
)
TENSOR_SEQUENCE_CODE_HELP = (
    """sequence_code of the tensor frame [taken from nef input if present]"""
)
TENSOR_RESIDUE_NAME_HELP = """\
    residue name of the tensor frame [taken from nef input if present, if you have multiple tensor frames the
    default will generally not work and you will need to provide chain codes [segids] or sequence  codes]"""


@export_app.command()
def rdcs(
    tensor_atom_names: List[str] = typer.Option(
        ["OO", "Z", "X", "Y"], "--tensor_atom_names", help=TENSOR_ATOMS_HELP
    ),
    tensor_chain_code: str = typer.Option(
        None, "--tensor-chain-code", help=TENSOR_CHAIN_CODE_HELP
    ),
    tensor_sequence_code: int = typer.Option(
        None, "--tensor-sequence-code", help=TENSOR_SEQUENCE_CODE_HELP
    ),
    tensor_residue_name: str = typer.Option(
        "ANI", "--tensor-residue-name", help=TENSOR_RESIDUE_NAME_HELP
    ),
    frame_selectors: List[str] = typer.Option(
        [],
        "-f",
        "--frame",
        help="selector for the rdc restraint frame to use, can be called multiple times and include wild cards",
    ),
):
    """- convert nef rdc restraints to xplor"""

    tensor_selector = SequenceResidue(
        tensor_chain_code, tensor_sequence_code, tensor_residue_name
    )

    entry = read_entry_from_file_or_stdin_or_exit_error(Path("-"))

    NEF_RDC_CATEGORY = "nef_rdc_restraint_list"

    rdc_frames = entry.get_saveframes_by_category(NEF_RDC_CATEGORY)

    if not frame_selectors:
        frame_selectors = ["*"]

    frames = select_frames_by_name(rdc_frames, frame_selectors)

    restraints = _rdc_restraints_from_frames(frames)

    for i, restraint_list in enumerate(restraints):
        restraints[i] = _translate_atom_names(restraint_list, "xplor")

    for restraint_list, frame in zip(restraints, frames):
        print(f"! restraints from frame {frame.name} in nef entry {entry.entry_id}")
        _print_restraints(restraint_list, tensor_selector, tensor_atom_names)
        print()


def _translate_atom_names(
    restraints: List[RdcRestraint], naming_scheme="iupac"
) -> List[RdcRestraint]:
    result = []
    for restraint in restraints:
        atom_1 = restraint.atom_1
        atom_2 = restraint.atom_2

        if atom_1.atom_name == "H":
            restraint.atom_1 = replace(atom_1, atom_name="HN")

        if atom_2.atom_name == "H":
            restraint.atom_1 = replace(atom_2, atom_name="HN")

        result.append(restraint)

    return result


def _rdc_restraints_from_frames(frames: List[Saveframe]):
    all_results = []
    for frame in frames:
        result = []
        all_results.append(result)

        for row in loop_row_dict_iter(frame.loops[0]):

            atom_1 = AtomLabel(
                SequenceResidue(
                    row["chain_code_1"],
                    int(row["sequence_code_1"]),
                    row["residue_name_1"],
                ),
                row["atom_name_1"],
            )
            atom_2 = AtomLabel(
                SequenceResidue(
                    row["chain_code_2"],
                    int(row["sequence_code_2"]),
                    row["residue_name_2"],
                ),
                row["atom_name_2"],
            )

            rdc = RdcRestraint(
                atom_1,
                atom_2,
                row["target_value"],
                row["target_value_uncertainty"],
            )

            result.append(rdc)

        result.sort()
    return all_results


def _print_restraints(
    restraints: List[RdcRestraint],
    tensor_frame_selector: SequenceResidue,
    tensor_atom_names: Tuple[str, str, str, str],
    value_uncertainty: float = 1.0,
):
    table = []

    for i, restraint in enumerate(restraints):
        origin = tensor_atom_names[0]

        tensor_chain_code = tensor_frame_selector.chain_code
        tensor_sequence_code = tensor_frame_selector.sequence_code
        tensor_residue_name = tensor_frame_selector.residue_name

        selectors = []
        for name, selector in zip(
            ("segid", "resid", "resn"),
            (tensor_chain_code, tensor_sequence_code, tensor_residue_name),
        ):
            if selector is not None:
                selectors.append(f"{name} {selector}")
        selectors = " and ".join(selectors).split()

        row = f"and name {origin} )".split()
        tensor_fields = ["assign", "(", *selectors, *row]
        table.append(tensor_fields)
        num_tensor_fields = len(tensor_fields)

        for name in tensor_atom_names[1:]:
            row = f"and name {name} )".split()
            tensor_fields = ["", "(", *selectors, *row]
            table.append(tensor_fields)
            num_tensor_fields = (
                len(tensor_fields)
                if len(tensor_fields) > num_tensor_fields
                else num_tensor_fields
            )

        atom_1 = restraint.atom_1
        chain_code_1 = atom_1.residue.chain_code
        sequence_code_1 = atom_1.residue.sequence_code
        residue_name_1 = atom_1.residue.residue_name
        atom_name_1 = atom_1.atom_name

        line = (
            f"( segid {chain_code_1} and resid {sequence_code_1}"
            f" and resname {residue_name_1} and name {atom_name_1} )"
        )
        row = line.split()
        row = ["", *row]
        num_row_fields = len(row)
        num_extra_spaces = num_tensor_fields - num_row_fields
        if num_extra_spaces < 0:
            num_extra_spaces = 0
        extra_spaces = [""] * num_extra_spaces
        row = [*row, *extra_spaces]
        table.append(row)

        atom_2 = restraint.atom_2
        chain_code_2 = atom_2.residue.chain_code
        sequence_code_2 = atom_2.residue.sequence_code
        residue_name_2 = atom_2.residue.residue_name
        atom_name_2 = atom_2.atom_name
        line = (
            f"( segid {chain_code_2} and resid {sequence_code_2}"
            f" and resname {residue_name_2} and name {atom_name_2} )"
        )
        row = line.split()
        row = ["", *row]
        num_row_fields = len(row)

        num_extra_spaces = num_tensor_fields - num_row_fields
        if num_extra_spaces < 0:
            num_extra_spaces = 0
        extra_spaces = [""] * num_extra_spaces
        row = [*row, *extra_spaces]

        table.append([*row, restraint.value, value_uncertainty])

        table.append([])

    print(tabulate.tabulate(table, tablefmt="plain", floatfmt="7.3f"))



from dataclasses import dataclass
from enum import auto
from typing import Dict, List, Optional, Union

from strenum import StrEnum


class Linking(StrEnum):
    START = (auto(),)
    MIDDLE = auto()
    END = auto()


@dataclass(frozen=True, order=True)
class Residue:
    chain_code: str
    sequence_code: Union[int, str]
    residue_name: str


@dataclass(frozen=True, order=True)
class SequenceResidue(Residue):

    is_cis: bool = False
    linking: Optional[Linking] = None
    variant: Optional[str] = None


# should contain a residue and have constructors?
@dataclass(frozen=True, order=True)
class AtomLabel:
    residue: Residue
    atom_name: str
    element: str = None
    isotope_number: int = None


@dataclass
class PeakAxis:
    atom_labels: List[AtomLabel]
    ppm: float
    merit: str
    # comment: str


@dataclass
class DistanceRestraint:
    atom_list_1: List[AtomLabel]
    atom_list_2: List[AtomLabel]

    target_distance: float
    distance_minus: float
    distance_plus: float


@dataclass
class DihedralRestraint:
    atom_1: AtomLabel
    atom_2: AtomLabel
    atom_3: AtomLabel
    atom_4: AtomLabel

    target_angle: float
    lower_limit: float
    upper_limit: float


@dataclass
class PeakValues:
    serial: int

    height: Optional[float] = None
    height_uncertainty: Optional[float] = None

    volume: Optional[float] = None
    volume_uncertainty: Optional[float] = None

    deleted: Optional[bool] = False
    comment: Optional[str] = ""

    width: Optional[float] = None  # HWHH ppm
    # bound: float

    # merit: Optional[str] = None

    # flag0: str


# assignment has a tuple of dimensions


@dataclass
class Assignments:
    assignments: Dict[str, List[AtomLabel]]


@dataclass
class Peak:
    id: int
    values: PeakValues

    # move these to axis_values?
    positions: Dict[str, float]

    # assignment has a list of one or more assignments
    # each Assignment will have one value for each axis this maybe be either
    # 0. a list with no AtomLabels - unassigned
    # 1. a list with a single AtomLabel -  this axis is definitively assigned
    # 2. a list with multiple AtomLabels - this axis has multiple putative assignments
    # Note if there are multiple unique assignments each of these is should be a top level
    # assignment of the peak
    assignments: List[Assignments]

    position_uncertainties: Optional[Dict[str, float]] = None


@dataclass
class PeakListData:
    num_axis: int
    axis_labels: List[str]
    # isotopes: List[str]
    data_set: str
    sweep_widths: List[float]
    spectrometer_frequencies: List[float]


# TODO: are axes indexed by names or by integer...
@dataclass
class PeakList:
    peak_list_data: PeakListData
    peaks: List[Dict[Union[int, str], Peak]]


@dataclass
class LineInfo:
    file_name: str
    line_no: int
    line: str


@dataclass(frozen=True, order=True)
class ShiftData:
    atom: AtomLabel
    value: float  # TODO: should be position
    value_uncertainty: Optional[float] = None  # TODO: should be position_uncertainty
    line_width: Optional[float] = None  # line width in Hz
    line_width_uncertainty: Optional[float] = None  # uncertainty of line width in Hz


@dataclass
class ShiftList:
    shifts: List[ShiftData]


@dataclass(order=True)
class RdcRestraint:
    atom_1: AtomLabel
    atom_2: AtomLabel
    value: float
    value_uncertainty: float
    weight: Optional[float] = None


@dataclass(frozen=True, order=True)
class NewPeak:

    shifts: List[ShiftData]

    id: Optional[int] = None
    height: Optional[float] = None
    height_uncertainty: Optional[float] = None
    volume: Optional[float] = None
    volume_uncertainty: Optional[float] = None


@dataclass
class DimensionInfo:
    axis_code: str  # this is the isotope code for us...
    axis_unit: Optional[str] = "ppm"
